import logging

import esphome.codegen as cg
from esphome.components import sensor, voltage_sampler
from esphome.components.esp32 import (
    get_esp32_variant,
    include_builtin_idf_component,
    require_adc_oneshot_iram,
)
from esphome.components.nrf52.const import AIN_TO_GPIO, EXTRA_ADC
from esphome.components.zephyr import (
    zephyr_add_overlay,
    zephyr_add_prj_conf,
    zephyr_add_user,
)
from esphome.config_helpers import filter_source_files_from_platform
import esphome.config_validation as cv
from esphome.const import (
    CONF_ATTENUATION,
    CONF_ID,
    CONF_NUMBER,
    CONF_PIN,
    CONF_RAW,
    DEVICE_CLASS_VOLTAGE,
    PLATFORM_NRF52,
    STATE_CLASS_MEASUREMENT,
    UNIT_VOLT,
    PlatformFramework,
)
from esphome.core import CORE
from esphome.types import ConfigType

from . import (
    ATTENUATION_MODES,
    ESP32_VARIANT_ADC1_PIN_TO_CHANNEL,
    ESP32_VARIANT_ADC2_PIN_TO_CHANNEL,
    SAMPLING_MODES,
    adc_ns,
    adc_unit_t,
    validate_adc_pin,
)

_LOGGER = logging.getLogger(__name__)

AUTO_LOAD = ["voltage_sampler"]

CONF_SAMPLES = "samples"
CONF_SAMPLING_MODE = "sampling_mode"


_attenuation = cv.enum(ATTENUATION_MODES, lower=True)
_sampling_mode = cv.enum(SAMPLING_MODES, lower=True)


def validate_config(config):
    if config[CONF_RAW] and config.get(CONF_ATTENUATION, None) == "auto":
        raise cv.Invalid("Automatic attenuation cannot be used when raw output is set")

    if config.get(CONF_ATTENUATION, None) == "auto" and config.get(CONF_SAMPLES, 1) > 1:
        raise cv.Invalid(
            "Automatic attenuation cannot be used when multisampling is set"
        )
    if config.get(CONF_ATTENUATION) == "11db":
        _LOGGER.warning(
            "`attenuation: 11db` is deprecated, use `attenuation: 12db` instead"
        )
        # Alter value here so `config` command prints the recommended change
        config[CONF_ATTENUATION] = _attenuation("12db")

    return config


def _require_adc_iram(config: ConfigType) -> ConfigType:
    """Register ADC oneshot IRAM requirement during config validation."""
    if CORE.is_esp32:
        require_adc_oneshot_iram()
    return config


ADCSensor = adc_ns.class_(
    "ADCSensor", sensor.Sensor, cg.PollingComponent, voltage_sampler.VoltageSampler
)

CONF_NRF_SAADC = "nrf_saadc"

adc_dt_spec = cg.global_ns.class_("adc_dt_spec")

CONFIG_SCHEMA = cv.All(
    sensor.sensor_schema(
        ADCSensor,
        unit_of_measurement=UNIT_VOLT,
        accuracy_decimals=2,
        device_class=DEVICE_CLASS_VOLTAGE,
        state_class=STATE_CLASS_MEASUREMENT,
    )
    .extend(
        {
            cv.Required(CONF_PIN): validate_adc_pin,
            cv.Optional(CONF_RAW, default=False): cv.boolean,
            cv.SplitDefault(CONF_ATTENUATION, esp32="0db"): cv.All(
                cv.only_on_esp32, _attenuation
            ),
            cv.OnlyWith(CONF_NRF_SAADC, PLATFORM_NRF52): cv.declare_id(adc_dt_spec),
            cv.Optional(CONF_SAMPLES, default=1): cv.int_range(min=1, max=255),
            cv.Optional(CONF_SAMPLING_MODE, default="avg"): _sampling_mode,
        }
    )
    .extend(cv.polling_component_schema("60s")),
    validate_config,
    _require_adc_iram,
)

CONF_ADC_CHANNEL_ID = "adc_channel_id"


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await sensor.register_sensor(var, config)

    if config[CONF_PIN] == "VCC":
        cg.add_define("USE_ADC_SENSOR_VCC")
    elif config[CONF_PIN] == "TEMPERATURE":
        cg.add(var.set_is_temperature())
    elif not CORE.is_nrf52 or config[CONF_PIN][CONF_NUMBER] not in EXTRA_ADC:
        pin = await cg.gpio_pin_expression(config[CONF_PIN])
        cg.add(var.set_pin(pin))

    cg.add(var.set_output_raw(config[CONF_RAW]))
    cg.add(var.set_sample_count(config[CONF_SAMPLES]))
    cg.add(var.set_sampling_mode(config[CONF_SAMPLING_MODE]))

    if CORE.is_esp32:
        # Re-enable ESP-IDF's ADC driver (excluded by default to save compile time)
        include_builtin_idf_component("esp_adc")

        if attenuation := config.get(CONF_ATTENUATION):
            if attenuation == "auto":
                cg.add(var.set_autorange(cg.global_ns.true))
            else:
                cg.add(var.set_attenuation(attenuation))

        variant = get_esp32_variant()
        pin_num = config[CONF_PIN][CONF_NUMBER]
        if (
            variant in ESP32_VARIANT_ADC1_PIN_TO_CHANNEL
            and pin_num in ESP32_VARIANT_ADC1_PIN_TO_CHANNEL[variant]
        ):
            chan = ESP32_VARIANT_ADC1_PIN_TO_CHANNEL[variant][pin_num]
            cg.add(var.set_channel(adc_unit_t.ADC_UNIT_1, chan))
        elif (
            variant in ESP32_VARIANT_ADC2_PIN_TO_CHANNEL
            and pin_num in ESP32_VARIANT_ADC2_PIN_TO_CHANNEL[variant]
        ):
            chan = ESP32_VARIANT_ADC2_PIN_TO_CHANNEL[variant][pin_num]
            cg.add(var.set_channel(adc_unit_t.ADC_UNIT_2, chan))

    elif CORE.is_nrf52:
        CORE.data.setdefault(CONF_ADC_CHANNEL_ID, 0)
        channel_id = CORE.data[CONF_ADC_CHANNEL_ID]
        CORE.data[CONF_ADC_CHANNEL_ID] = channel_id + 1
        zephyr_add_prj_conf("ADC", True)
        nrf_saadc = config[CONF_NRF_SAADC]
        rhs = cg.RawExpression(
            f"ADC_DT_SPEC_GET_BY_IDX(DT_PATH(zephyr_user), {channel_id})"
        )
        adc = cg.new_Pvariable(nrf_saadc, rhs)
        cg.add(var.set_adc_channel(adc))
        gain = "ADC_GAIN_1_6"
        pin_number = config[CONF_PIN][CONF_NUMBER]
        if pin_number == "VDDHDIV5":
            gain = "ADC_GAIN_1_2"
        if isinstance(pin_number, int):
            GPIO_TO_AIN = {v: k for k, v in AIN_TO_GPIO.items()}
            pin_number = GPIO_TO_AIN[pin_number]
        zephyr_add_user("io-channels", f"<&adc {channel_id}>")
        zephyr_add_overlay(
            f"""
                &adc {{
                    #address-cells = <1>;
                    #size-cells = <0>;

                    channel@{channel_id} {{
                        reg = <{channel_id}>;
                        zephyr,gain = "{gain}";
                        zephyr,reference = "ADC_REF_INTERNAL";
                        zephyr,acquisition-time = <ADC_ACQ_TIME_DEFAULT>;
                        zephyr,input-positive = <NRF_SAADC_{pin_number}>;
                        zephyr,resolution = <14>;
                        zephyr,oversampling = <8>;
                    }};
                }};
            """
        )


FILTER_SOURCE_FILES = filter_source_files_from_platform(
    {
        "adc_sensor_esp32.cpp": {
            PlatformFramework.ESP32_ARDUINO,
            PlatformFramework.ESP32_IDF,
        },
        "adc_sensor_esp8266.cpp": {PlatformFramework.ESP8266_ARDUINO},
        "adc_sensor_rp2.cpp": {PlatformFramework.RP2_ARDUINO},
        "adc_sensor_libretiny.cpp": {
            PlatformFramework.BK72XX_ARDUINO,
            PlatformFramework.RTL87XX_ARDUINO,
            PlatformFramework.LN882X_ARDUINO,
        },
        "adc_sensor_zephyr.cpp": {PlatformFramework.NRF52_ZEPHYR},
    }
)



"""
LVGL 9.4 Canvas Widget Implementation

This module implements the canvas widget for LVGL 9.4. Key changes from LVGL 8.4:

1. Buffer allocation:
   - LV_IMG_CF_TRUE_COLOR → LV_COLOR_FORMAT_RGB565
   - LV_IMG_CF_TRUE_COLOR_ALPHA → LV_COLOR_FORMAT_ARGB8888
   - LV_CANVAS_BUF_SIZE_TRUE_COLOR → LV_CANVAS_BUF_SIZE(w, h, bpp, stride)

2. Drawing API:
   - All lv_canvas_draw_* functions removed
   - Use layer-based drawing: lv_canvas_init_layer() / lv_canvas_finish_layer()
   - Draw using low-level lv_draw_* functions (rect, line, arc, image, label)

3. Pixel operations:
   - lv_canvas_set_px_color + lv_canvas_set_px_opa → lv_canvas_set_px(color, opa)
"""

from esphome import automation, codegen as cg, config_validation as cv
from esphome.components.display_menu_base import CONF_LABEL
from esphome.const import (
    CONF_COLOR,
    CONF_HEIGHT,
    CONF_ID,
    CONF_TEXT,
    CONF_WIDTH,
    CONF_X,
    CONF_Y,
)
from esphome.cpp_types import FixedVector

from ..automation import action_to_code
from ..defines import (
    CONF_END_ANGLE,
    CONF_MAIN,
    CONF_OPA,
    CONF_PIVOT_X,
    CONF_PIVOT_Y,
    CONF_POINTS,
    CONF_RADIUS,
    CONF_SRC,
    CONF_START_ANGLE,
    addr,
    literal,
)
from ..lv_validation import (
    lv_angle_degrees,
    lv_bool,
    lv_color,
    lv_image,
    lv_text,
    opacity,
    pixels,
    pixels_or_percent,
    size,
)
from ..lvcode import LocalVariable, lv, lv_assign, lv_expr
from ..schemas import STYLE_PROPS, TEXT_SCHEMA, point_schema, remap_property
from ..types import LvType, ObjUpdateAction, lv_point_precise_t
from . import Widget, WidgetType, get_widgets
from .img import CONF_IMAGE

CONF_CANVAS = "canvas"
CONF_BUFFER_ID = "buffer_id"
CONF_MAX_WIDTH = "max_width"
CONF_TRANSPARENT = "transparent"
CONF_DRAW_BUF_ID = "draw_buf_id"

lv_canvas_t = LvType("lv_canvas_t")
lv_draw_buf_t = LvType("lv_draw_buf_t")


class CanvasType(WidgetType):
    def __init__(self):
        super().__init__(
            CONF_CANVAS,
            lv_canvas_t,
            (CONF_MAIN,),
            cv.Schema(
                {
                    cv.Required(CONF_WIDTH): size,
                    cv.Required(CONF_HEIGHT): size,
                    cv.Optional(CONF_TRANSPARENT, default=False): cv.boolean,
                    cv.GenerateID(CONF_DRAW_BUF_ID): cv.declare_id(lv_draw_buf_t),
                }
            ),
            modify_schema={},
        )

    def get_uses(self):
        return CONF_IMAGE, CONF_LABEL

    async def to_code(self, w: Widget, config):
        width = config[CONF_WIDTH]
        height = config[CONF_HEIGHT]
        # LVGL 9.4: Use LV_COLOR_FORMAT instead of LV_IMG_CF
        # RGB565 is 16-bit (2 bytes per pixel), ARGB8888 is 32-bit (4 bytes per pixel)
        if config[CONF_TRANSPARENT]:
            color_format = "LV_COLOR_FORMAT_ARGB8888"
        else:
            color_format = "LV_COLOR_FORMAT_NATIVE"

        # LVGL 9.4: LV_CANVAS_BUF_SIZE(width, height, bits_per_pixel, stride)
        # stride is 0 for default (width * bytes_per_pixel)
        draw_buf = cg.new_Pvariable(config[CONF_DRAW_BUF_ID])
        buf_size = literal(f"LV_DRAW_BUF_SIZE({width}, {height}, {color_format})")
        lv.draw_buf_init(
            draw_buf,
            width,
            height,
            literal(color_format),
            0,
            lv_expr.malloc_core(buf_size),
            literal(buf_size),
        )
        lv.draw_buf_set_flag(draw_buf, literal("LV_IMAGE_FLAGS_MODIFIABLE"))
        lv.canvas_set_draw_buf(w.obj, draw_buf)


CanvasType()


@automation.register_action(
    "lvgl.canvas.fill",
    ObjUpdateAction,
    cv.Schema(
        {
            cv.GenerateID(CONF_ID): cv.use_id(lv_canvas_t),
            cv.Required(CONF_COLOR): lv_color,
            cv.Optional(CONF_OPA, default="COVER"): opacity,
        },
    ),
    synchronous=True,
)
async def canvas_fill(config, action_id, template_arg, args):
    widget = await get_widgets(config)
    color = await lv_color.process(config[CONF_COLOR])
    opa = await opacity.process(config[CONF_OPA])

    async def do_fill(w: Widget):
        lv.canvas_fill_bg(w.obj, color, opa)

    return await action_to_code(widget, do_fill, action_id, template_arg, args, config)


@automation.register_action(
    "lvgl.canvas.set_pixels",
    ObjUpdateAction,
    cv.Schema(
        {
            cv.GenerateID(CONF_ID): cv.use_id(lv_canvas_t),
            cv.Required(CONF_COLOR): lv_color,
            cv.Optional(CONF_OPA, default="COVER"): opacity,
            cv.Required(CONF_POINTS): cv.ensure_list(point_schema),
        },
    ),
    synchronous=True,
)
async def canvas_set_pixel(config, action_id, template_arg, args):
    widget = await get_widgets(config)
    color = await lv_color.process(config[CONF_COLOR])
    opa = await opacity.process(config.get(CONF_OPA), "COVER")
    points = [
        (
            await pixels.process(p[CONF_X]),
            await pixels.process(p[CONF_Y]),
        )
        for p in config[CONF_POINTS]
    ]

    async def do_set_pixels(w: Widget):
        # LVGL 9.4: lv_canvas_set_px combines color and opacity
        # Could optimize this for lambda values
        for point in points:
            x, y = point
            lv.canvas_set_px(w.obj, x, y, color, opa)

    return await action_to_code(
        widget, do_set_pixels, action_id, template_arg, args, config
    )


DRAW_SCHEMA = {
    cv.GenerateID(CONF_ID): cv.use_id(lv_canvas_t),
    cv.Required(CONF_X): pixels,
    cv.Required(CONF_Y): pixels,
}
DRAW_OPA_SCHEMA = {
    **DRAW_SCHEMA,
    cv.Optional(CONF_OPA): opacity,
}


async def draw_to_code(config, dsc_type, props, do_draw, action_id, template_arg, args):
    widget = await get_widgets(config)
    x = await pixels.process(config.get(CONF_X))
    y = await pixels.process(config.get(CONF_Y))

    async def action_func(w: Widget):
        # LVGL 9.4: Create a layer for drawing on canvas
        with LocalVariable("layer", "lv_layer_t", modifier="") as layer:
            lv.canvas_init_layer(w.obj, addr(layer))
            with LocalVariable("dsc", f"lv_draw_{dsc_type}_dsc_t", modifier="") as dsc:
                lv.call(f"draw_{dsc_type}_dsc_init", addr(dsc))
                if CONF_OPA in config:
                    opa = await opacity.process(config[CONF_OPA])
                    lv_assign(dsc.opa, opa)
                for prop, validator in props.items():
                    if prop in config:
                        value = await validator.process(config[prop])
                        mapped_prop = remap_property(prop)
                        lv_assign(getattr(dsc, mapped_prop), value)
                await do_draw(addr(layer), x, y, dsc)
            lv.canvas_finish_layer(w.obj, addr(layer))

    return await action_to_code(
        widget, action_func, action_id, template_arg, args, config
    )


RECT_PROPS = {
    p: STYLE_PROPS[p]
    for p in (
        "radius",
        "bg_opa",
        "bg_color",
        "bg_grad",
        "border_color",
        "border_width",
        "border_opa",
        "outline_color",
        "outline_width",
        "outline_pad",
        "outline_opa",
        "shadow_color",
        "shadow_width",
        "shadow_offset_x",
        "shadow_offset_y",
        "shadow_ofs_x",
        "shadow_ofs_y",
        "shadow_spread",
        "shadow_opa",
    )
}


def _draw_line(layer, dsc, points):
    # LVGL 9.4: Use lv_draw_line for each line segment
    with (
        LocalVariable(
            "points", FixedVector.template(lv_point_precise_t), points, modifier=""
        ) as points_var,
        LocalVariable("i", "uint32_t", literal("0"), modifier="") as i,
    ):
        # Draw lines between consecutive points
        lv.append(
            cg.RawStatement(f"for ({i} = 0; {i} != {points_var}.size() - 1; {i}++) {{")
        )
        lv_assign(dsc.p1, points_var[i])
        lv_assign(dsc.p2, points_var[i + 1])
        lv.draw_line(layer, addr(dsc))
        lv.append(cg.RawStatement("}"))


@automation.register_action(
    "lvgl.canvas.draw_rectangle",
    ObjUpdateAction,
    cv.Schema(
        {
            **DRAW_OPA_SCHEMA,
            cv.Required(CONF_WIDTH): cv.templatable(cv.int_),
            cv.Required(CONF_HEIGHT): cv.templatable(cv.int_),
            **{cv.Optional(prop): STYLE_PROPS[prop] for prop in RECT_PROPS},
        }
    ),
    synchronous=True,
)
async def canvas_draw_rect(config, action_id, template_arg, args):
    width = await pixels.process(config[CONF_WIDTH])
    height = await pixels.process(config[CONF_HEIGHT])

    async def do_draw_rect(layer, x, y, dsc):
        # LVGL 9.4: Use lv_draw_rect with area
        with LocalVariable("area", "lv_area_t", modifier="") as area:
            lv_assign(area.x1, x)
            lv_assign(area.y1, y)
            lv_assign(area.x2, literal(f"{x} + {width} - 1"))
            lv_assign(area.y2, literal(f"{y} + {height} - 1"))
            lv.draw_rect(layer, addr(dsc), addr(area))

    return await draw_to_code(
        config, "rect", RECT_PROPS, do_draw_rect, action_id, template_arg, args
    )


TEXT_PROPS = {
    p: STYLE_PROPS[f"text_{p}"]
    for p in (
        "font",
        "color",
        # "sel_color",
        # "sel_bg_color",
        "line_space",
        "letter_space",
        "align",
        "decor",
    )
}


@automation.register_action(
    "lvgl.canvas.draw_text",
    ObjUpdateAction,
    cv.Schema(
        {
            **TEXT_SCHEMA,
            **DRAW_OPA_SCHEMA,
            cv.Required(CONF_MAX_WIDTH): cv.templatable(cv.int_),
            **{cv.Optional(prop): STYLE_PROPS[f"text_{prop}"] for prop in TEXT_PROPS},
        },
    ),
    synchronous=True,
)
async def canvas_draw_text(config, action_id, template_arg, args):
    text = await lv_text.process(config[CONF_TEXT])
    max_width = await pixels.process(config[CONF_MAX_WIDTH])

    async def do_draw_text(layer, x, y, dsc):
        # LVGL 9.4: Use lv_draw_label with area and hint
        with LocalVariable("area", "lv_area_t", modifier="") as area:
            lv_assign(area.x1, x)
            lv_assign(area.y1, y)
            lv_assign(area.x2, literal(f"{x} + {max_width} - 1"))
            lv_assign(area.y2, literal(f"{y} + LV_COORD_MAX"))
            lv_assign(dsc.text, text)
            lv.draw_label(layer, addr(dsc), addr(area))

    return await draw_to_code(
        config, "label", TEXT_PROPS, do_draw_text, action_id, template_arg, args
    )


IMG_PROPS = (
    "angle",
    "rotation",
    "scale_x",
    "scale_y",
    "skew_x",
    "skew_y",
    "scale",
    "zoom",
    "recolor",
    "recolor_opa",
    "opa",
)


def _scale_map(config):
    config = {remap_property(p): v for p, v in config.items()}
    if "scale" in config and {"scale_x", "scale_y"} & config.keys():
        raise cv.Invalid("Cannot specify both scale and scale_x/scale_y")
    if "scale" in config:
        config.update({"scale_x": config["scale"], "scale_y": config["scale"]})
        del config["scale"]
    return config


def _get_prop_validator(prop):
    return STYLE_PROPS.get(
        f"transform_{remap_property(prop, False)}"
    ) or STYLE_PROPS.get(prop)


def _prop_validator(prop):
    def validator(value):
        return _get_prop_validator(prop)(value)

    return validator


@automation.register_action(
    "lvgl.canvas.draw_image",
    ObjUpdateAction,
    cv.Schema(
        {
            **DRAW_OPA_SCHEMA,
            cv.Required(CONF_SRC): lv_image,
            cv.Optional(CONF_PIVOT_X, default=0): pixels,
            cv.Optional(CONF_PIVOT_Y, default=0): pixels,
            **{cv.Optional(prop): _prop_validator(prop) for prop in IMG_PROPS},
        }
    ).add_extra(_scale_map),
    synchronous=True,
)
async def canvas_draw_image(config, action_id, template_arg, args):
    src = await lv_image.process(config[CONF_SRC])
    pivot_x = await pixels.process(config[CONF_PIVOT_X])
    pivot_y = await pixels.process(config[CONF_PIVOT_Y])

    async def do_draw_image(layer, x, y, dsc):
        # LVGL 9.4: Use lv_draw_image with area
        lv_assign(dsc.src, src.get_lv_image_dsc())
        if pivot_x or pivot_y:
            # pylint :disable=no-member
            lv_assign(dsc.pivot.x, pivot_x)
            lv_assign(dsc.pivot.y, pivot_y)
        with LocalVariable("area", "lv_area_t", modifier="") as area:
            lv_assign(area.x1, x)
            lv_assign(area.y1, y)
            # Image size will be determined from the image descriptor
            lv_assign(area.x2, x)
            lv_assign(area.y2, y)
            lv.draw_image(layer, addr(dsc), addr(area))

    return await draw_to_code(
        config,
        "image",
        {prop: _get_prop_validator(prop) for prop in IMG_PROPS},
        do_draw_image,
        action_id,
        template_arg,
        args,
    )


LINE_PROPS = {
    "width": STYLE_PROPS["line_width"],
    "color": STYLE_PROPS["line_color"],
    "dash-width": STYLE_PROPS["line_dash_width"],
    "dash-gap": STYLE_PROPS["line_dash_gap"],
    "round_start": lv_bool,
    "round_end": lv_bool,
}


def _validate_points(config):
    for index, point in enumerate(config[CONF_POINTS]):
        if not all(isinstance(p, int) for p in point.values()):
            raise cv.Invalid("Points must be integers", path=[CONF_POINTS, index])
    return config


@automation.register_action(
    "lvgl.canvas.draw_line",
    ObjUpdateAction,
    cv.Schema(
        {
            cv.GenerateID(CONF_ID): cv.use_id(lv_canvas_t),
            cv.Optional(CONF_OPA): opacity,
            cv.Required(CONF_POINTS): cv.ensure_list(point_schema),
            **{cv.Optional(prop): validator for prop, validator in LINE_PROPS.items()},
        }
    ).add_extra(_validate_points),
    synchronous=True,
)
async def canvas_draw_line(config, action_id, template_arg, args):
    points = [
        [
            await pixels.process(p[CONF_X]),
            await pixels.process(p[CONF_Y]),
        ]
        for p in config[CONF_POINTS]
    ]

    async def do_draw_line(layer, _x, _y, dsc):
        _draw_line(layer, dsc, points)

    return await draw_to_code(
        config, "line", LINE_PROPS, do_draw_line, action_id, template_arg, args
    )


@automation.register_action(
    "lvgl.canvas.draw_polygon",
    ObjUpdateAction,
    cv.Schema(
        {
            cv.GenerateID(CONF_ID): cv.use_id(lv_canvas_t),
            cv.Required(CONF_POINTS): cv.ensure_list(point_schema),
            **{cv.Optional(prop): STYLE_PROPS[prop] for prop in RECT_PROPS},
        },
    ).add_extra(_validate_points),
    synchronous=True,
)
async def canvas_draw_polygon(config, action_id, template_arg, args):
    points = [
        [
            await pixels_or_percent.process(p[CONF_X]),
            await pixels_or_percent.process(p[CONF_Y]),
        ]
        for p in config[CONF_POINTS]
    ]
    # Close the polygon
    points.append(points[0])

    async def do_draw_polygon(layer, x, y, dsc):
        # LVGL 9.4: Draw polygon using line drawing in a closed path
        # Note: This draws outline only. For filled polygons, would need different approach
        # Convert rect descriptor to line descriptor for polygon outline
        with LocalVariable("line_dsc", "lv_draw_line_dsc_t", modifier="") as line_dsc:
            lv.draw_line_dsc_init(addr(line_dsc))
            # Copy border properties from rect descriptor to line descriptor
            lv_assign(line_dsc.color, dsc.border_color)
            lv_assign(line_dsc.width, dsc.border_width)
            lv_assign(line_dsc.opa, dsc.border_opa)
            _draw_line(layer, line_dsc, points)

    return await draw_to_code(
        config, "rect", RECT_PROPS, do_draw_polygon, action_id, template_arg, args
    )


ARC_PROPS = {
    "width": STYLE_PROPS["arc_width"],
    "color": STYLE_PROPS["arc_color"],
    "rounded": STYLE_PROPS["arc_rounded"],
}


@automation.register_action(
    "lvgl.canvas.draw_arc",
    ObjUpdateAction,
    cv.Schema(
        {
            **DRAW_OPA_SCHEMA,
            cv.Required(CONF_RADIUS): pixels,
            cv.Required(CONF_START_ANGLE): lv_angle_degrees,
            cv.Required(CONF_END_ANGLE): lv_angle_degrees,
            **{cv.Optional(prop): validator for prop, validator in ARC_PROPS.items()},
        }
    ),
    synchronous=True,
)
async def canvas_draw_arc(config, action_id, template_arg, args):
    radius = await size.process(config[CONF_RADIUS])
    start_angle = await lv_angle_degrees.process(config[CONF_START_ANGLE])
    end_angle = await lv_angle_degrees.process(config[CONF_END_ANGLE])

    async def do_draw_arc(layer, x, y, dsc):
        # LVGL 9.4: Use lv_draw_arc with center point
        lv_assign(dsc.center.x, x)
        lv_assign(dsc.center.y, y)
        lv_assign(dsc.start_angle, start_angle)
        lv_assign(dsc.end_angle, end_angle)
        lv_assign(dsc.radius, radius)
        lv.draw_arc(layer, addr(dsc))

    return await draw_to_code(
        config, "arc", ARC_PROPS, do_draw_arc, action_id, template_arg, args
    )



import esphome.config_validation as cv

from ..defines import CONF_ARC_LENGTH, CONF_INDICATOR, CONF_MAIN, CONF_SPIN_TIME
from ..lv_validation import lv_angle_degrees, lv_milliseconds
from ..lvcode import lv
from ..types import LvType
from . import Widget, WidgetType
from .arc import CONF_ARC

CONF_SPINNER = "spinner"

SPINNER_SCHEMA = cv.Schema(
    {
        cv.Optional(CONF_ARC_LENGTH, default=200): cv.All(
            lv_angle_degrees, cv.int_range(min=0, max=360)
        ),
        cv.Optional(CONF_SPIN_TIME, default="2s"): lv_milliseconds,
    }
)


class SpinnerType(WidgetType):
    def __init__(self):
        super().__init__(
            CONF_SPINNER,
            LvType("lv_spinner_t"),
            (CONF_MAIN, CONF_INDICATOR),
            SPINNER_SCHEMA,
        )

    async def to_code(self, w: Widget, config):
        spin_time = await lv_milliseconds.process(config.get(CONF_SPIN_TIME))
        arc_length = int(config[CONF_ARC_LENGTH])
        if arc_length < 180:
            arc_length += 180
        lv.spinner_set_anim_params(w.obj, spin_time, arc_length)

    def get_uses(self):
        return (CONF_ARC,)


spinner_spec = SpinnerType()



import esphome.config_validation as cv
from esphome.const import CONF_HEIGHT, CONF_WIDTH

from ..defines import CONF_CONTAINER, CONF_MAIN, CONF_SCROLLBAR
from ..lv_validation import size
from ..types import lv_obj_t
from . import WidgetType

CONTAINER_SCHEMA = cv.Schema(
    {
        cv.Optional(CONF_HEIGHT, default="100%"): size,
        cv.Optional(CONF_WIDTH, default="100%"): size,
    }
)


class ContainerType(WidgetType):
    """
    A simple container widget that can hold other widgets and which defaults to a 100% size.
    Made from an obj with all styles removed
    """

    def __init__(self):
        super().__init__(
            CONF_CONTAINER,
            lv_obj_t,
            (CONF_MAIN, CONF_SCROLLBAR),
            schema=CONTAINER_SCHEMA,
            modify_schema={},
            lv_name=CONF_CONTAINER,
        )
        self.styles = {}


container_spec = ContainerType()



import base64
from pathlib import Path
import re
import secrets

import requests
from ruamel.yaml import YAML

from esphome import git
import esphome.codegen as cg
from esphome.components.packages import validate_source_shorthand
import esphome.config_validation as cv
from esphome.const import CONF_ESPHOME, CONF_PROJECT, CONF_REF, CONF_WIFI
import esphome.final_validate as fv
from esphome.yaml_util import dump

dashboard_import_ns = cg.esphome_ns.namespace("dashboard_import")

# payload is in `esphomelib` mdns record, which only exists if api
# is enabled
DEPENDENCIES = ["api"]
CODEOWNERS = ["@esphome/core"]


def validate_import_url(value):
    value = cv.string_strict(value)
    value = cv.Length(max=255)(value)
    validate_source_shorthand(value)
    return value


def validate_full_url(config):
    if not config[CONF_IMPORT_FULL_CONFIG]:
        return config
    source = validate_source_shorthand(config[CONF_PACKAGE_IMPORT_URL])
    if CONF_REF not in source:
        raise cv.Invalid(
            "Must specify a ref (branch or tag) to import from when importing full config"
        )
    return config


CONF_PACKAGE_IMPORT_URL = "package_import_url"
CONF_IMPORT_FULL_CONFIG = "import_full_config"

CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.Required(CONF_PACKAGE_IMPORT_URL): validate_import_url,
            cv.Optional(CONF_IMPORT_FULL_CONFIG, default=False): cv.boolean,
        }
    ),
    validate_full_url,
)


def _final_validate(config):
    full_config = fv.full_config.get()[CONF_ESPHOME]
    if CONF_PROJECT not in full_config:
        raise cv.Invalid(
            "Dashboard import requires the `esphome` -> `project` information to be provided."
        )


FINAL_VALIDATE_SCHEMA = _final_validate

WIFI_CONFIG = """

wifi:
  ssid: !secret wifi_ssid
  password: !secret wifi_password
"""


async def to_code(config):
    cg.add_define("USE_DASHBOARD_IMPORT")
    url = config[CONF_PACKAGE_IMPORT_URL]
    if config[CONF_IMPORT_FULL_CONFIG]:
        url += "?full_config"
    cg.add(dashboard_import_ns.set_package_import_url(cg.FlashStringLiteral(url)))


def import_config(
    path: str,
    name: str,
    friendly_name: str | None,
    project_name: str,
    import_url: str,
    network: str = CONF_WIFI,
    encryption: bool = False,
) -> None:
    """Materialise a dashboard-imported device's YAML on disk.

    Used by:
    - device-builder (esphome/device-builder) — called from the
      ``devices/import`` WS handler to seed the YAML for an adopted
      factory firmware. Coordinate before changing the kwargs or the
      generated YAML's top-level keys; both consumers depend on the
      output shape (``esphome.name`` / ``packages:`` import url) to
      route subsequent compile + flash operations.
    """
    p = Path(path)

    if p.exists():
        raise FileExistsError

    git_file = git.GitFile.from_shorthand(import_url)

    if git_file.query and "full_config" in git_file.query:
        url = git_file.raw_url
        try:
            req = requests.get(url, timeout=30)
            req.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise ValueError(f"Error while fetching {url}: {e}") from e

        contents = req.text
        yaml = YAML()
        loaded_yaml = yaml.load(contents)
        if (
            "name_add_mac_suffix" in loaded_yaml["esphome"]
            and loaded_yaml["esphome"]["name_add_mac_suffix"]
        ):
            loaded_yaml["esphome"]["name_add_mac_suffix"] = False
            name_val = loaded_yaml["esphome"]["name"]
            sub_pattern = re.compile(r"\$\{?([a-zA-Z-_]+)\}?")
            if match := sub_pattern.match(name_val):
                name_sub = match.group(1)
                if name_sub in loaded_yaml["substitutions"]:
                    loaded_yaml["substitutions"][name_sub] = name
                else:
                    raise ValueError(
                        f"Name substitution {name_sub} not found in substitutions"
                    )
            else:
                loaded_yaml["esphome"]["name"] = name
            if friendly_name is not None:
                friendly_name_val = loaded_yaml["esphome"]["friendly_name"]
                if match := sub_pattern.match(friendly_name_val):
                    friendly_name_sub = match.group(1)
                    if friendly_name_sub in loaded_yaml["substitutions"]:
                        loaded_yaml["substitutions"][friendly_name_sub] = friendly_name
                    else:
                        raise ValueError(
                            f"Friendly name substitution {friendly_name_sub} not found in substitutions"
                        )
                else:
                    loaded_yaml["esphome"]["friendly_name"] = friendly_name

            with p.open("w", encoding="utf8") as f:
                yaml.dump(loaded_yaml, f)
        else:
            with p.open("w", encoding="utf8") as f:
                f.write(contents)

    else:
        substitutions = {"name": name}
        esphome_core = {"name": "${name}", "name_add_mac_suffix": False}
        if friendly_name:
            substitutions["friendly_name"] = friendly_name
            esphome_core["friendly_name"] = "${friendly_name}"
        config = {
            "substitutions": substitutions,
            "packages": {project_name: import_url},
            "esphome": esphome_core,
        }
        if encryption:
            noise_psk = secrets.token_bytes(32)
            key = base64.b64encode(noise_psk).decode()
            config["api"] = {"encryption": {"key": key}}

        output = dump(config)

        if network == CONF_WIFI:
            output += WIFI_CONFIG

        p.write_text(output, encoding="utf8")



from esphome import automation, codegen as cg, config_validation as cv
from esphome.automation import Trigger, build_automation
from esphome.config_validation import COMPONENT_SCHEMA
from esphome.const import (
    CONF_ACCELERATION,
    CONF_DURATION,
    CONF_FROM,
    CONF_ID,
    CONF_ON_START,
    CONF_TIMING,
    CONF_TO,
    CONF_TRIGGER_ID,
    CONF_TYPE,
    CONF_WEIGHT,
)
from esphome.cpp_generator import MockObj, TemplateArguments

from ..const import CONF_LOOP
from .defines import (
    CONF_AUTO_START,
    CONF_LVGL_ID,
    CONF_ON_STOP,
    CONF_WIDGETS,
    LValidator,
    add_define,
    literal,
)
from .lv_validation import (
    color,
    get_component_colors,
    lv_color,
    lv_milliseconds,
    lv_positive_float,
    lv_zero_to_one_float,
)
from .lvcode import LVGL_COMP_ARG, LambdaContext, LvglComponent, lv_add
from .schemas import STYLE_PROPS
from .types import LvAnimation, LvglAction, lv_color_t, lv_coord_t, lv_obj_t, lvgl_ns
from .widgets import get_widgets

LvAnimationTimingRoundTrip = lvgl_ns.class_("LvAnimationTimingRoundTrip")
LvAnimationTimingEaseInOut = lvgl_ns.class_("LvAnimationTimingEaseInOut")

CONF_BOUNCE = "bounce"


def timing_class(name, extras=None):
    # Convert config option to camel case
    cls_name = "LvAnimationTiming" + "".join([w.capitalize() for w in name.split("_")])
    cls = lvgl_ns.class_(cls_name)
    schema = cv.Schema({cv.GenerateID(): cv.declare_id(cls)})
    if extras:
        schema = schema.extend(extras)
    return name, schema


# TODO - currently the order of arguments to timing classes is expected to be alphabetical, but this is not enforced.
# It would be better to have a more robust way of passing arguments to the timing classes.
TIMING_SCHEMA = cv.maybe_simple_value(
    cv.typed_schema(
        dict(
            [
                timing_class("round_trip"),
                timing_class(
                    "ease_in_out",
                    {cv.Optional(CONF_WEIGHT, default=2.0): lv_positive_float},
                ),
                timing_class(
                    "gravity",
                    {
                        cv.Optional(CONF_ACCELERATION, default=0.5): lv_positive_float,
                        cv.Optional(CONF_BOUNCE, default=0.5): lv_zero_to_one_float,
                    },
                ),
            ]
        ),
        default_type="ease_in_out",
    ),
    key=CONF_TYPE,
)

CONF_START_DELAY = "start_delay"


class LiteralColorValidator(LValidator):
    def __init__(self):
        super().__init__(
            color, lv_color_t, retmapper=get_component_colors, animatable=True
        )

    def __call__(self, value):
        if isinstance(value, cv.Lambda):
            raise cv.Invalid(
                "An animated color may not be set with a lambda, only a literal color value."
            )
        return super().__call__(value)


literal_color = LiteralColorValidator()


def from_to(validator):
    return cv.Schema(
        {
            cv.Required(CONF_FROM): validator,
            cv.Required(CONF_TO): validator,
        }
    )


# Colors can only be animated between constants, not lambdas.
def map_v(validator):
    if validator == lv_color:
        return literal_color
    return validator


ANIMABLE_STYLES = {
    k: map_v(v)
    for k, v in STYLE_PROPS.items()
    if isinstance(v, LValidator) and v.animatable
}

ANIMATION_SCHEMA = cv.Schema(
    {
        cv.Optional(CONF_AUTO_START, default=False): cv.boolean,
        cv.Optional(CONF_LOOP, default=False): cv.boolean,
        cv.Optional(CONF_DURATION, default="5s"): lv_milliseconds,
        cv.Optional(CONF_START_DELAY, default="0s"): lv_milliseconds,
        cv.Optional(CONF_TIMING, default=[]): cv.ensure_list(TIMING_SCHEMA),
        cv.Required(CONF_ID): cv.declare_id(LvAnimation),
        cv.Optional(CONF_ON_START): automation.validate_automation(
            {
                cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(Trigger.template()),
            }
        ),
        cv.Optional(CONF_ON_STOP): automation.validate_automation(
            {
                cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(Trigger.template()),
            }
        ),
        cv.Required(CONF_WIDGETS): cv.ensure_list(
            cv.Schema(
                {
                    cv.Required(CONF_ID): cv.use_id(lv_obj_t),
                }
            ).extend({cv.Optional(k): from_to(v) for k, v in ANIMABLE_STYLES.items()})
        ),
    }
).extend(COMPONENT_SCHEMA)


async def _process_arg(validator, arg) -> list:
    # from/to values are evaluated at animation start with no arguments, so the
    # generated lambda must be parameterless rather than inheriting the enclosing
    # update-callback's `values` parameter.
    value = await validator.process(arg, args=[], raw_lambda=True)
    value = list(value) if isinstance(value, tuple) else [value]
    return [literal(f"TemplatableValue<lv_coord_t>({v})") for v in value]


async def animations_to_code(config):
    for animation in config:
        add_define("USE_LVGL_ANIMATION")
        widgets = animation[CONF_WIDGETS]
        async with LambdaContext(
            [(lv_coord_t.operator("const").operator("ptr"), "values")]
        ) as ctx:
            froms = []
            tos = []
            for widget in widgets:
                w = (await get_widgets(widget))[0]
                props = [(k, v) for k, v in widget.items() if k in ANIMABLE_STYLES]
                for prop, value_range in props:
                    # prop is the style property, value_range is a dict with from: and to: values
                    validator = ANIMABLE_STYLES[prop]
                    from_value = await _process_arg(validator, value_range[CONF_FROM])
                    to_value = await _process_arg(validator, value_range[CONF_TO])
                    index = len(froms)
                    if len(from_value) == 1:
                        value = f"values[{index}]"
                    else:
                        value = f"lv_color_make(values[{index}+0], values[{index}+1], values[{index}+2])"
                    w.set_style(prop, literal(value), 0)
                    # The value arrays are extended by 1 item for scalar properties, 3 for colors
                    froms.extend(from_value)
                    tos.extend(to_value)

        data_size = len(froms)
        loop = animation[CONF_LOOP]
        start_delay = await lv_milliseconds.process(animation.get(CONF_START_DELAY))
        var = cg.new_Pvariable(
            animation[CONF_ID],
            TemplateArguments(data_size, animation[CONF_AUTO_START]),
            await ctx.get_lambda(),
            froms,
            tos,
        )
        for timing in animation[CONF_TIMING]:
            timing_id = timing[CONF_ID]
            args = sorted(
                [(k, v) for k, v in timing.items() if k not in [CONF_ID, CONF_TYPE]]
            )
            args = [v for k, v in args]
            timing_var = cg.new_Pvariable(timing_id, *args)
            cg.add(var.add_timing(timing_var))

        if start_delay:
            cg.add(var.set_start_delay(start_delay))
        if loop:
            cg.add(var.set_loop(loop))
        cg.add(
            var.set_duration(await lv_milliseconds.process(animation[CONF_DURATION]))
        )
        await cg.register_component(var, animation)


async def add_animation_triggers(config):
    async def add_triggers(animation: MockObj, event: str, config: dict) -> None:
        for conf in config:
            trigger = cg.new_Pvariable(conf[CONF_TRIGGER_ID])
            await build_automation(trigger, [], conf)
            async with LambdaContext([]) as context:
                lv_add(trigger.trigger())
            lv_add(
                getattr(
                    animation,
                    f"add_{event}_callback",
                )(await context.get_lambda())
            )

    for animation in config:
        var = await cg.get_variable(animation[CONF_ID])
        await add_triggers(var, CONF_ON_START, animation.get(CONF_ON_START, []))
        await add_triggers(var, CONF_ON_STOP, animation.get(CONF_ON_STOP, []))


@automation.register_action(
    "lvgl.animation.start",
    LvglAction,
    cv.maybe_simple_value(
        {
            cv.Required(CONF_ID): cv.ensure_list(cv.use_id(LvAnimation)),
            cv.GenerateID(CONF_LVGL_ID): cv.use_id(LvglComponent),
            cv.Optional(CONF_DURATION): lv_milliseconds,
            cv.Optional(CONF_START_DELAY): lv_milliseconds,
            cv.Optional(CONF_LOOP): cv.boolean,
        },
        key=CONF_ID,
    ),
    synchronous=True,
)
async def start_animation(config, action_id, template_arg, args):
    animations = config[CONF_ID]
    loop = config.get(CONF_LOOP)
    async with LambdaContext(LVGL_COMP_ARG, where=action_id) as context:
        for animation in animations:
            anim_var = await cg.get_variable(animation)
            if loop is not None:
                context.add(anim_var.set_loop(loop))
            if (duration := config.get(CONF_DURATION)) is not None:
                context.add(
                    anim_var.set_duration(await lv_milliseconds.process(duration))
                )
            if (start_delay := config.get(CONF_START_DELAY)) is not None:
                context.add(
                    anim_var.set_start_delay(await lv_milliseconds.process(start_delay))
                )
            context.add(anim_var.start())
    var = cg.new_Pvariable(action_id, template_arg, await context.get_lambda())
    await cg.register_parented(var, config[CONF_LVGL_ID])
    return var


@automation.register_action(
    "lvgl.animation.stop",
    LvglAction,
    cv.maybe_simple_value(
        {
            cv.Required(CONF_ID): cv.ensure_list(cv.use_id(LvAnimation)),
            cv.GenerateID(CONF_LVGL_ID): cv.use_id(LvglComponent),
        },
        key=CONF_ID,
    ),
    synchronous=True,
)
async def stop_animation(config, action_id, template_arg, args):
    animations = config[CONF_ID]
    async with LambdaContext(LVGL_COMP_ARG, where=action_id) as context:
        for animation in animations:
            anim_var = await cg.get_variable(animation)
            context.add(anim_var.stop())
    var = cg.new_Pvariable(action_id, template_arg, await context.get_lambda())
    await cg.register_parented(var, config[CONF_LVGL_ID])
    return var



from __future__ import annotations

import asyncio
from datetime import datetime
import importlib
import logging
from typing import TYPE_CHECKING, Any
import warnings

# Suppress protobuf version warnings
with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore", category=UserWarning, message=".*Protobuf gencode version.*"
    )
    from aioesphomeapi import APIClient, parse_log_message
    from aioesphomeapi.log_runner import async_run

import contextlib

from esphome.const import CONF_KEY, CONF_PORT, __version__
from esphome.core import CORE, EsphomeError
from esphome.util import safe_print

from . import CONF_ENCRYPTION

if TYPE_CHECKING:
    from aioesphomeapi.api_pb2 import (
        SubscribeLogsResponse,  # pylint: disable=no-name-in-module
    )


_LOGGER = logging.getLogger(__name__)


class _LogLineProcessor:
    """Feeds incoming log lines to the stack-trace decoder.

    Two responsibilities beyond just calling the decoder:
    1. Catch EsphomeError. on_log runs inside an asyncio protocol
       callback; if an exception escapes, the loop tears the transport
       down with "Fatal error: protocol.data_received() call failed."
       and ReconnectLogic immediately reconnects, the device replays
       the same crash trace, and we loop forever.
    2. Disable decoding after the first failure. _decode_pc shells out
       to PlatformIO via _run_idedata, which is expensive; a single
       crash dump can contain many PC/BT lines and we don't want to
       retry the failing subprocess for each one.
    """

    def __init__(self, config: dict[str, Any], platform_handler: Any | None) -> None:
        self._config = config
        self._platform_handler = platform_handler
        self._decode_enabled = True
        self.backtrace_state = False

    def process_line(self, raw_line: str) -> None:
        if not self._decode_enabled:
            return
        try:
            if self._platform_handler is not None:
                self.backtrace_state = self._platform_handler(
                    self._config, raw_line, self.backtrace_state
                )
        except EsphomeError as exc:
            self._decode_enabled = False
            self.backtrace_state = False
            # _run_idedata raises EsphomeError with no message; fall back
            # to a generic explanation when str(exc) is empty.
            detail = str(exc) or "build artifacts not found locally"
            _LOGGER.warning(
                "Crash trace decoding unavailable: %s. "
                "Run 'esphome compile' for this device to enable PC decoding.",
                detail,
            )


async def async_run_logs(
    config: dict[str, Any],
    addresses: list[str],
    subscribe_states: bool = True,
) -> None:
    """Run the logs command in the event loop."""
    conf = config["api"]
    name = config["esphome"]["name"]
    port: int = int(conf[CONF_PORT])
    noise_psk: str | None = None
    if (encryption := conf.get(CONF_ENCRYPTION)) and (key := encryption.get(CONF_KEY)):
        noise_psk = key

    if len(addresses) == 1:
        _LOGGER.info("Starting log output from %s using esphome API", addresses[0])
    else:
        _LOGGER.info(
            "Starting log output from %s using esphome API", " or ".join(addresses)
        )

    cli = APIClient(
        addresses[0],  # Primary address for compatibility
        port,
        "",  # Password auth removed in 2026.1.0
        client_info=f"ESPHome Logs {__version__}",
        noise_psk=noise_psk,
        addresses=addresses,  # Pass all addresses for automatic retry
        provide_time=False,
    )

    # Try platform-specific stacktrace handler first, fall back to generic
    platform_process_stacktrace = None
    try:
        module = importlib.import_module("esphome.components." + CORE.target_platform)
        platform_process_stacktrace = module.process_stacktrace
    except (AttributeError, ImportError):
        _LOGGER.info(
            'Stacktrace analysis is unavailable: no compatible analyzer found for target platform "%s".',
            CORE.target_platform,
        )

    processor = _LogLineProcessor(config, platform_process_stacktrace)

    def on_log(msg: SubscribeLogsResponse) -> None:
        """Handle a new log message."""
        time_ = datetime.now().astimezone()
        message: bytes = msg.message
        text = message.decode("utf8", "backslashreplace")
        nanoseconds = time_.microsecond // 1000
        timestamp = (
            f"[{time_.hour:02}:{time_.minute:02}:{time_.second:02}.{nanoseconds:03}]"
        )
        for parsed_msg in parse_log_message(text, timestamp):
            # safe_print handles the dashboard \033 escaping and falls back
            # to backslashreplace encoding on stdouts that can't represent
            # the wifi signal-bar block characters (Windows redirected
            # cp1252 pipe).
            safe_print(parsed_msg)
        for raw_line in text.splitlines():
            processor.process_line(raw_line)

    # Safe to fall back to plaintext here only for this diagnostics use
    # case: the stream is one-way from device to client, and this code
    # never accepts commands or acts on any message the device sends.
    # An on-path attacker could still both inject fabricated log lines
    # and passively read the device's log output (and any state data
    # delivered when subscribe_states is enabled), so this does lose
    # confidentiality as well as authentication/integrity. That tradeoff
    # is acceptable for operator-visible logs, which aioesphomeapi also
    # warns may come from an unverified device. Never mirror this opt-in
    # for any connection that sends data to the device or uses Home
    # Assistant actions.
    stop = await async_run(
        cli,
        on_log,
        name=name,
        subscribe_states=subscribe_states,
        allow_plaintext_fallback=True,
        # A top-level ``deep_sleep:`` block means the device is only awake
        # briefly; cap the reconnect backoff so a wake window is not missed.
        deep_sleep="deep_sleep" in config,
    )
    try:
        await asyncio.Event().wait()
    finally:
        await stop()


def run_logs(
    config: dict[str, Any],
    addresses: list[str],
    subscribe_states: bool = True,
) -> None:
    """Run the logs command."""
    with contextlib.suppress(KeyboardInterrupt):
        asyncio.run(
            async_run_logs(config, addresses, subscribe_states=subscribe_states)
        )



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_HUMIDITY,
    CONF_OUTDOOR_TEMPERATURE,
    CONF_POWER,
    DEVICE_CLASS_CURRENT,
    DEVICE_CLASS_FREQUENCY,
    DEVICE_CLASS_HUMIDITY,
    DEVICE_CLASS_POWER,
    DEVICE_CLASS_TEMPERATURE,
    ENTITY_CATEGORY_DIAGNOSTIC,
    ICON_CURRENT_AC,
    ICON_FLASH,
    ICON_GAUGE,
    ICON_HEATING_COIL,
    ICON_PULSE,
    ICON_THERMOMETER,
    ICON_WATER_PERCENT,
    ICON_WEATHER_WINDY,
    STATE_CLASS_MEASUREMENT,
    UNIT_AMPERE,
    UNIT_CELSIUS,
    UNIT_HERTZ,
    UNIT_PERCENT,
    UNIT_WATT,
)

from ..climate import CONF_HAIER_ID, HonClimate

CODEOWNERS = ["@paveldn"]
SensorTypeEnum = HonClimate.enum("SubSensorType", True)

# Haier sensors
CONF_COMPRESSOR_CURRENT = "compressor_current"
CONF_COMPRESSOR_FREQUENCY = "compressor_frequency"
CONF_EXPANSION_VALVE_OPEN_DEGREE = "expansion_valve_open_degree"
CONF_INDOOR_COIL_TEMPERATURE = "indoor_coil_temperature"
CONF_OUTDOOR_COIL_TEMPERATURE = "outdoor_coil_temperature"
CONF_OUTDOOR_DEFROST_TEMPERATURE = "outdoor_defrost_temperature"
CONF_OUTDOOR_IN_AIR_TEMPERATURE = "outdoor_in_air_temperature"
CONF_OUTDOOR_OUT_AIR_TEMPERATURE = "outdoor_out_air_temperature"

# Additional icons
ICON_SNOWFLAKE_THERMOMETER = "mdi:snowflake-thermometer"

SENSOR_TYPES = {
    CONF_COMPRESSOR_CURRENT: sensor.sensor_schema(
        unit_of_measurement=UNIT_AMPERE,
        icon=ICON_CURRENT_AC,
        accuracy_decimals=1,
        device_class=DEVICE_CLASS_CURRENT,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_COMPRESSOR_FREQUENCY: sensor.sensor_schema(
        unit_of_measurement=UNIT_HERTZ,
        icon=ICON_PULSE,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_FREQUENCY,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_EXPANSION_VALVE_OPEN_DEGREE: sensor.sensor_schema(
        unit_of_measurement=UNIT_PERCENT,
        icon=ICON_GAUGE,
        accuracy_decimals=2,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_HUMIDITY: sensor.sensor_schema(
        unit_of_measurement=UNIT_PERCENT,
        icon=ICON_WATER_PERCENT,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_HUMIDITY,
        state_class=STATE_CLASS_MEASUREMENT,
    ),
    CONF_INDOOR_COIL_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_HEATING_COIL,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_OUTDOOR_COIL_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_HEATING_COIL,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_OUTDOOR_DEFROST_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_SNOWFLAKE_THERMOMETER,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_OUTDOOR_IN_AIR_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_WEATHER_WINDY,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_OUTDOOR_OUT_AIR_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_WEATHER_WINDY,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
    CONF_OUTDOOR_TEMPERATURE: sensor.sensor_schema(
        unit_of_measurement=UNIT_CELSIUS,
        icon=ICON_THERMOMETER,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_TEMPERATURE,
        state_class=STATE_CLASS_MEASUREMENT,
    ),
    CONF_POWER: sensor.sensor_schema(
        unit_of_measurement=UNIT_WATT,
        icon=ICON_FLASH,
        accuracy_decimals=0,
        device_class=DEVICE_CLASS_POWER,
        state_class=STATE_CLASS_MEASUREMENT,
        entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
    ),
}

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_HAIER_ID): cv.use_id(HonClimate),
    }
).extend({cv.Optional(type_): schema for type_, schema in SENSOR_TYPES.items()})


async def to_code(config):
    paren = await cg.get_variable(config[CONF_HAIER_ID])

    for type_ in SENSOR_TYPES:
        if conf := config.get(type_):
            sens = await sensor.new_sensor(conf)
            sensor_type = getattr(SensorTypeEnum, type_.upper())
            cg.add(paren.set_sub_sensor(sensor_type, sens))



import esphome.codegen as cg
from esphome.components import button
import esphome.config_validation as cv
from esphome.const import CONF_ID
from esphome.core import CORE

DEPENDENCIES = ["network"]


def AUTO_LOAD():
    if CORE.is_esp8266 or CORE.is_rp2:
        return []
    return ["socket"]


CONF_TARGET_MAC_ADDRESS = "target_mac_address"

wake_on_lan_ns = cg.esphome_ns.namespace("wake_on_lan")

WakeOnLanButton = wake_on_lan_ns.class_("WakeOnLanButton", button.Button, cg.Component)

CONFIG_SCHEMA = (
    button.button_schema(WakeOnLanButton)
    .extend(cv.COMPONENT_SCHEMA)
    .extend(
        {
            cv.Required(CONF_TARGET_MAC_ADDRESS): cv.mac_address,
        }
    )
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    cg.add(var.set_macaddr(*config[CONF_TARGET_MAC_ADDRESS].parts))
    await cg.register_component(var, config)
    await button.register_button(var, config)



from esphome import automation
import esphome.codegen as cg
from esphome.components import i2c, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_ACTUAL_GAIN,
    CONF_ACTUAL_INTEGRATION_TIME,
    CONF_AMBIENT_LIGHT,
    CONF_AUTO_MODE,
    CONF_FULL_SPECTRUM_COUNTS,
    CONF_GAIN,
    CONF_GLASS_ATTENUATION_FACTOR,
    CONF_ID,
    CONF_INTEGRATION_TIME,
    CONF_NAME,
    CONF_REPEAT,
    CONF_TYPE,
    DEVICE_CLASS_ILLUMINANCE,
    ICON_BRIGHTNESS_5,
    ICON_BRIGHTNESS_6,
    ICON_TIMER,
    STATE_CLASS_MEASUREMENT,
    UNIT_LUX,
    UNIT_MILLISECOND,
)

CODEOWNERS = ["@latonita"]
DEPENDENCIES = ["i2c"]

CONF_INFRARED_COUNTS = "infrared_counts"
CONF_ON_PS_HIGH_THRESHOLD = "on_ps_high_threshold"
CONF_ON_PS_LOW_THRESHOLD = "on_ps_low_threshold"
CONF_PS_COOLDOWN = "ps_cooldown"
CONF_PS_COUNTS = "ps_counts"
CONF_PS_GAIN = "ps_gain"
CONF_PS_HIGH_THRESHOLD = "ps_high_threshold"
CONF_PS_LOW_THRESHOLD = "ps_low_threshold"
ICON_BRIGHTNESS_7 = "mdi:brightness-7"
ICON_GAIN = "mdi:multiplication"
ICON_PROXIMITY = "mdi:hand-wave-outline"
UNIT_COUNTS = "#"

ltr_als_ps_ns = cg.esphome_ns.namespace("ltr_als_ps")

LTRAlsPsComponent = ltr_als_ps_ns.class_(
    "LTRAlsPsComponent", cg.PollingComponent, i2c.I2CDevice
)

LtrType = ltr_als_ps_ns.enum("LtrType")
LTR_TYPES = {
    "ALS": LtrType.LTR_TYPE_ALS_ONLY,
    "PS": LtrType.LTR_TYPE_PS_ONLY,
    "ALS_PS": LtrType.LTR_TYPE_ALS_AND_PS,
}

AlsGain = ltr_als_ps_ns.enum("AlsGain")
ALS_GAINS = {
    "1X": AlsGain.GAIN_1,
    "2X": AlsGain.GAIN_2,
    "4X": AlsGain.GAIN_4,
    "8X": AlsGain.GAIN_8,
    "48X": AlsGain.GAIN_48,
    "96X": AlsGain.GAIN_96,
}

IntegrationTime = ltr_als_ps_ns.enum("IntegrationTime")
INTEGRATION_TIMES = {
    50: IntegrationTime.INTEGRATION_TIME_50MS,
    100: IntegrationTime.INTEGRATION_TIME_100MS,
    150: IntegrationTime.INTEGRATION_TIME_150MS,
    200: IntegrationTime.INTEGRATION_TIME_200MS,
    250: IntegrationTime.INTEGRATION_TIME_250MS,
    300: IntegrationTime.INTEGRATION_TIME_300MS,
    350: IntegrationTime.INTEGRATION_TIME_350MS,
    400: IntegrationTime.INTEGRATION_TIME_400MS,
}

MeasurementRepeatRate = ltr_als_ps_ns.enum("MeasurementRepeatRate")
MEASUREMENT_REPEAT_RATES = {
    50: MeasurementRepeatRate.REPEAT_RATE_50MS,
    100: MeasurementRepeatRate.REPEAT_RATE_100MS,
    200: MeasurementRepeatRate.REPEAT_RATE_200MS,
    500: MeasurementRepeatRate.REPEAT_RATE_500MS,
    1000: MeasurementRepeatRate.REPEAT_RATE_1000MS,
    2000: MeasurementRepeatRate.REPEAT_RATE_2000MS,
}

PsGain = ltr_als_ps_ns.enum("PsGain")
PS_GAINS = {
    "16X": PsGain.PS_GAIN_16,
    "32X": PsGain.PS_GAIN_32,
    "64X": PsGain.PS_GAIN_64,
}


def validate_integration_time(value):
    value = cv.positive_time_period_milliseconds(value).total_milliseconds
    return cv.enum(INTEGRATION_TIMES, int=True)(value)


def validate_repeat_rate(value):
    value = cv.positive_time_period_milliseconds(value).total_milliseconds
    return cv.enum(MEASUREMENT_REPEAT_RATES, int=True)(value)


def validate_time_and_repeat_rate(config):
    integraton_time = config[CONF_INTEGRATION_TIME]
    repeat_rate = config[CONF_REPEAT]
    if integraton_time > repeat_rate:
        raise cv.Invalid(
            f"Measurement repeat rate ({repeat_rate}ms) shall be greater or equal to integration time ({integraton_time}ms)"
        )
    return config


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(LTRAlsPsComponent),
            cv.Optional(CONF_TYPE, default="ALS_PS"): cv.enum(LTR_TYPES, upper=True),
            cv.Optional(CONF_AUTO_MODE, default=True): cv.boolean,
            cv.Optional(CONF_GAIN, default="1X"): cv.enum(ALS_GAINS, upper=True),
            cv.Optional(
                CONF_INTEGRATION_TIME, default="100ms"
            ): validate_integration_time,
            cv.Optional(CONF_REPEAT, default="500ms"): validate_repeat_rate,
            cv.Optional(CONF_GLASS_ATTENUATION_FACTOR, default=1.0): cv.float_range(
                min=1.0
            ),
            cv.Optional(
                CONF_PS_COOLDOWN, default="5s"
            ): cv.positive_time_period_seconds,
            cv.Optional(CONF_PS_GAIN, default="16X"): cv.enum(PS_GAINS, upper=True),
            cv.Optional(CONF_PS_HIGH_THRESHOLD, default=65535): cv.int_range(
                min=0, max=65535
            ),
            cv.Optional(CONF_PS_LOW_THRESHOLD, default=0): cv.int_range(
                min=0, max=65535
            ),
            cv.Optional(CONF_ON_PS_HIGH_THRESHOLD): automation.validate_automation({}),
            cv.Optional(CONF_ON_PS_LOW_THRESHOLD): automation.validate_automation({}),
            cv.Optional(CONF_AMBIENT_LIGHT): cv.maybe_simple_value(
                sensor.sensor_schema(
                    unit_of_measurement=UNIT_LUX,
                    icon=ICON_BRIGHTNESS_6,
                    accuracy_decimals=1,
                    device_class=DEVICE_CLASS_ILLUMINANCE,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
            cv.Optional(CONF_INFRARED_COUNTS): cv.maybe_simple_value(
                sensor.sensor_schema(
                    unit_of_measurement=UNIT_COUNTS,
                    icon=ICON_BRIGHTNESS_5,
                    accuracy_decimals=0,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
            cv.Optional(CONF_FULL_SPECTRUM_COUNTS): cv.maybe_simple_value(
                sensor.sensor_schema(
                    unit_of_measurement=UNIT_COUNTS,
                    icon=ICON_BRIGHTNESS_7,
                    accuracy_decimals=0,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
            cv.Optional(CONF_PS_COUNTS): cv.maybe_simple_value(
                sensor.sensor_schema(
                    unit_of_measurement=UNIT_COUNTS,
                    icon=ICON_PROXIMITY,
                    accuracy_decimals=0,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
            cv.Optional(CONF_ACTUAL_GAIN): cv.maybe_simple_value(
                sensor.sensor_schema(
                    icon=ICON_GAIN,
                    accuracy_decimals=0,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
            cv.Optional(CONF_ACTUAL_INTEGRATION_TIME): cv.maybe_simple_value(
                sensor.sensor_schema(
                    unit_of_measurement=UNIT_MILLISECOND,
                    icon=ICON_TIMER,
                    accuracy_decimals=0,
                    state_class=STATE_CLASS_MEASUREMENT,
                ),
                key=CONF_NAME,
            ),
        }
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(i2c.i2c_device_schema(0x29)),
    validate_time_and_repeat_rate,
)


_CALLBACK_AUTOMATIONS = (
    automation.CallbackAutomation(
        CONF_ON_PS_HIGH_THRESHOLD, "add_on_ps_high_trigger_callback"
    ),
    automation.CallbackAutomation(
        CONF_ON_PS_LOW_THRESHOLD, "add_on_ps_low_trigger_callback"
    ),
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)

    if als_config := config.get(CONF_AMBIENT_LIGHT):
        sens = await sensor.new_sensor(als_config)
        cg.add(var.set_ambient_light_sensor(sens))

    if infrared_cnt_config := config.get(CONF_INFRARED_COUNTS):
        sens = await sensor.new_sensor(infrared_cnt_config)
        cg.add(var.set_infrared_counts_sensor(sens))

    if full_spect_cnt_config := config.get(CONF_FULL_SPECTRUM_COUNTS):
        sens = await sensor.new_sensor(full_spect_cnt_config)
        cg.add(var.set_full_spectrum_counts_sensor(sens))

    if act_gain_config := config.get(CONF_ACTUAL_GAIN):
        sens = await sensor.new_sensor(act_gain_config)
        cg.add(var.set_actual_gain_sensor(sens))

    if act_itime_config := config.get(CONF_ACTUAL_INTEGRATION_TIME):
        sens = await sensor.new_sensor(act_itime_config)
        cg.add(var.set_actual_integration_time_sensor(sens))

    if prox_cnt_config := config.get(CONF_PS_COUNTS):
        sens = await sensor.new_sensor(prox_cnt_config)
        cg.add(var.set_proximity_counts_sensor(sens))

    await automation.build_callback_automations(var, config, _CALLBACK_AUTOMATIONS)

    cg.add(var.set_ltr_type(config[CONF_TYPE]))

    cg.add(var.set_als_auto_mode(config[CONF_AUTO_MODE]))
    cg.add(var.set_als_gain(config[CONF_GAIN]))
    cg.add(var.set_als_integration_time(config[CONF_INTEGRATION_TIME]))
    cg.add(var.set_als_meas_repeat_rate(config[CONF_REPEAT]))
    cg.add(var.set_als_glass_attenuation_factor(config[CONF_GLASS_ATTENUATION_FACTOR]))

    cg.add(var.set_ps_cooldown_time_s(config[CONF_PS_COOLDOWN]))
    cg.add(var.set_ps_gain(config[CONF_PS_GAIN]))
    cg.add(var.set_ps_high_threshold(config[CONF_PS_HIGH_THRESHOLD]))
    cg.add(var.set_ps_low_threshold(config[CONF_PS_LOW_THRESHOLD]))



import hashlib
import json
import logging
from pathlib import Path
import re
from urllib.parse import urljoin

from esphome import automation, external_files, git
from esphome.automation import register_action, register_condition
from esphome.bundle import add_bundle_file
import esphome.codegen as cg
from esphome.components import esp32, microphone, ota, psram
from esphome.components.http_request import validate_url
import esphome.config_validation as cv
from esphome.const import (
    CONF_FILE,
    CONF_ID,
    CONF_INTERNAL,
    CONF_MICROPHONE,
    CONF_MODEL,
    CONF_PASSWORD,
    CONF_PATH,
    CONF_RAW_DATA_ID,
    CONF_REF,
    CONF_REFRESH,
    CONF_TASK_STACK_IN_PSRAM,
    CONF_TYPE,
    CONF_URL,
    CONF_USERNAME,
    TYPE_GIT,
    TYPE_LOCAL,
)
from esphome.core import CORE, HexInt
from esphome.types import ConfigType

_LOGGER = logging.getLogger(__name__)

AUTO_LOAD = ["ring_buffer"]
CODEOWNERS = ["@kahrendt", "@jesserockz"]
DEPENDENCIES = ["microphone"]

DOMAIN = "micro_wake_word"


CONF_FEATURE_STEP_SIZE = "feature_step_size"
CONF_MODELS = "models"
CONF_ON_WAKE_WORD_DETECTED = "on_wake_word_detected"
CONF_PROBABILITY_CUTOFF = "probability_cutoff"
CONF_SLIDING_WINDOW_AVERAGE_SIZE = "sliding_window_average_size"
CONF_SLIDING_WINDOW_SIZE = "sliding_window_size"
CONF_STOP_AFTER_DETECTION = "stop_after_detection"
CONF_TENSOR_ARENA_SIZE = "tensor_arena_size"
CONF_VAD = "vad"

TYPE_HTTP = "http"

micro_wake_word_ns = cg.esphome_ns.namespace("micro_wake_word")

MicroWakeWord = micro_wake_word_ns.class_("MicroWakeWord", cg.Component)

DisableModelAction = micro_wake_word_ns.class_("DisableModelAction", automation.Action)
EnableModelAction = micro_wake_word_ns.class_("EnableModelAction", automation.Action)
StartAction = micro_wake_word_ns.class_("StartAction", automation.Action)
StopAction = micro_wake_word_ns.class_("StopAction", automation.Action)

ModelIsEnabledCondition = micro_wake_word_ns.class_(
    "ModelIsEnabledCondition", automation.Condition
)
IsRunningCondition = micro_wake_word_ns.class_(
    "IsRunningCondition", automation.Condition
)

WakeWordModel = micro_wake_word_ns.class_("WakeWordModel")


def _validate_json_filename(value):
    value = cv.string(value)
    if not value.endswith(".json"):
        raise cv.Invalid("Manifest filename must end with .json")
    return value


def _process_git_source(config):
    repo_dir, _ = git.clone_or_update(
        url=config[CONF_URL],
        ref=config.get(CONF_REF),
        refresh=config[CONF_REFRESH],
        domain=DOMAIN,
        username=config.get(CONF_USERNAME),
        password=config.get(CONF_PASSWORD),
    )

    if not (repo_dir / config[CONF_FILE]).exists():
        raise cv.Invalid("File does not exist in repository")

    return config


CV_GIT_SCHEMA = cv.GIT_SCHEMA
if isinstance(CV_GIT_SCHEMA, dict):
    CV_GIT_SCHEMA = cv.Schema(CV_GIT_SCHEMA)

GIT_SCHEMA = cv.All(
    CV_GIT_SCHEMA.extend(
        {
            cv.Required(CONF_FILE): _validate_json_filename,
            cv.Optional(CONF_REFRESH, default="1d"): cv.All(
                cv.string, cv.source_refresh
            ),
        }
    ),
    _process_git_source,
)


KEY_AUTHOR = "author"
KEY_MICRO = "micro"
KEY_MINIMUM_ESPHOME_VERSION = "minimum_esphome_version"
KEY_TRAINED_LANGUAGES = "trained_languages"
KEY_VERSION = "version"
KEY_WAKE_WORD = "wake_word"
KEY_WEBSITE = "website"

MANIFEST_SCHEMA_V1 = cv.Schema(
    {
        cv.Required(CONF_TYPE): "micro",
        cv.Required(KEY_WAKE_WORD): cv.string,
        cv.Required(KEY_AUTHOR): cv.string,
        cv.Required(KEY_WEBSITE): cv.url,
        cv.Required(KEY_VERSION): cv.All(cv.int_, 1),
        cv.Required(CONF_MODEL): cv.string,
        cv.Required(KEY_MICRO): cv.Schema(
            {
                cv.Required(CONF_PROBABILITY_CUTOFF): cv.float_,
                cv.Required(CONF_SLIDING_WINDOW_AVERAGE_SIZE): cv.positive_int,
                cv.Optional(KEY_MINIMUM_ESPHOME_VERSION): cv.All(
                    cv.version_number, cv.validate_esphome_version
                ),
            }
        ),
    }
)

MANIFEST_SCHEMA_V2 = cv.Schema(
    {
        cv.Required(CONF_TYPE): "micro",
        cv.Required(CONF_MODEL): cv.string,
        cv.Required(KEY_AUTHOR): cv.string,
        cv.Required(KEY_VERSION): cv.All(cv.int_, 2),
        cv.Required(KEY_WAKE_WORD): cv.string,
        cv.Required(KEY_TRAINED_LANGUAGES): cv.ensure_list(cv.string),
        cv.Optional(KEY_WEBSITE): cv.url,
        cv.Required(KEY_MICRO): cv.Schema(
            {
                cv.Required(CONF_FEATURE_STEP_SIZE): cv.int_range(min=0, max=30),
                cv.Required(CONF_TENSOR_ARENA_SIZE): cv.int_,
                cv.Required(CONF_PROBABILITY_CUTOFF): cv.float_,
                cv.Required(CONF_SLIDING_WINDOW_SIZE): cv.positive_int,
                cv.Required(KEY_MINIMUM_ESPHOME_VERSION): cv.All(
                    cv.version_number, cv.validate_esphome_version
                ),
            }
        ),
    }
)


def _compute_local_file_path(config: dict) -> Path:
    url = config[CONF_URL]
    h = hashlib.new("sha256")
    h.update(url.encode())
    key = h.hexdigest()[:8]
    base_dir = external_files.compute_local_file_dir(DOMAIN)
    return base_dir / key


def _convert_manifest_v1_to_v2(v1_manifest):
    v2_manifest = v1_manifest.copy()

    v2_manifest[KEY_VERSION] = 2
    v2_manifest[KEY_MICRO][CONF_SLIDING_WINDOW_SIZE] = v1_manifest[KEY_MICRO][
        CONF_SLIDING_WINDOW_AVERAGE_SIZE
    ]
    del v2_manifest[KEY_MICRO][CONF_SLIDING_WINDOW_AVERAGE_SIZE]

    # Original Inception-based V1 manifest models require a minimum of 45672 bytes
    v2_manifest[KEY_MICRO][CONF_TENSOR_ARENA_SIZE] = 45672
    # Original Inception-based V1 manifest models use a 20 ms feature step size
    v2_manifest[KEY_MICRO][CONF_FEATURE_STEP_SIZE] = 20
    # Original Inception-based V1 manifest models were trained only on TTS English samples
    v2_manifest[KEY_TRAINED_LANGUAGES] = ["en"]

    return v2_manifest


def _validate_manifest_version(manifest_data):
    if manifest_version := manifest_data.get(KEY_VERSION):
        if manifest_version == 1:
            try:
                MANIFEST_SCHEMA_V1(manifest_data)
            except cv.Invalid as e:
                raise cv.Invalid(f"Invalid manifest file: {e}") from e
        elif manifest_version == 2:
            try:
                MANIFEST_SCHEMA_V2(manifest_data)
            except cv.Invalid as e:
                raise cv.Invalid(f"Invalid manifest file: {e}") from e
        else:
            raise cv.Invalid("Invalid manifest version")
    else:
        raise cv.Invalid("Invalid manifest file, missing 'version' key")


HTTP_SCHEMA = cv.Schema(
    {
        # validate_url only accepts http(s); the shorthand validator relies
        # on this branch rejecting git shorthands ("github://...") so they
        # fall through to the git branch.
        cv.Required(CONF_URL): validate_url,
    }
)


def _register_local_model_file(config: ConfigType) -> ConfigType:
    """Register the model file that the manifest points to, so bundles include it.

    The manifest names its model file relative to itself, so that path never appears
    in the YAML and bundle discovery cannot find it on its own.

    Problems with the manifest are logged and ignored here rather than raised. Loading
    the manifest later reports them with better messages, and raising would be
    swallowed by the shorthand validator, which then reports a confusing error about a
    missing file in a git repository. Logging keeps the skipped registration
    diagnosable if the manifest is only briefly unreadable, since the bundle would
    then be built without the model file.
    """
    manifest_path: Path = config[CONF_PATH]
    try:
        manifest = json.loads(manifest_path.read_text(encoding="utf-8"))
        model = manifest[CONF_MODEL]
    except (OSError, ValueError, KeyError, TypeError) as err:
        _LOGGER.debug("Not registering a model file from %s: %s", manifest_path, err)
        return config
    if not isinstance(model, str):
        _LOGGER.debug(
            "Not registering a model file from %s: 'model' is %s, expected a string",
            manifest_path,
            type(model).__name__,
        )
        return config
    add_bundle_file(manifest_path.parent / model)
    return config


LOCAL_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.Required(CONF_PATH): cv.All(_validate_json_filename, cv.file_),
        }
    ),
    _register_local_model_file,
)


# Bare model names in the official model repository ("okay_nabu"). Must not
# overlap with local paths, http(s) urls, or git shorthands
# ("github://user/repo/file.json@ref"), which the shorthand validator tries
# next; anything containing "/", ":" or "@" is not a model name.
_MODEL_NAME_RE = re.compile(r"[A-Za-z0-9_.-]+")


def _validate_source_model_name(value):
    if not isinstance(value, str):
        raise cv.Invalid("Model name must be a string")

    if value.endswith(".json"):
        raise cv.Invalid("Model name must not end with .json")

    if not _MODEL_NAME_RE.fullmatch(value):
        raise cv.Invalid("Model name may only contain letters, numbers, . _ -")

    return MODEL_SOURCE_SCHEMA(
        {
            CONF_TYPE: TYPE_HTTP,
            CONF_URL: f"https://github.com/esphome/micro-wake-word-models/raw/main/models/v2/{value}.json",
        }
    )


def _validate_source_shorthand(value):
    if not isinstance(value, str):
        raise cv.Invalid("Shorthand only for strings")

    try:  # Test for model name
        return _validate_source_model_name(value)
    except cv.Invalid:
        pass

    try:  # Test for local path
        return MODEL_SOURCE_SCHEMA({CONF_TYPE: TYPE_LOCAL, CONF_PATH: value})
    except cv.Invalid:
        pass

    try:  # Test for http url
        return MODEL_SOURCE_SCHEMA({CONF_TYPE: TYPE_HTTP, CONF_URL: value})
    except cv.Invalid:
        pass

    git_file = git.GitFile.from_shorthand(value)

    conf = {
        CONF_TYPE: TYPE_GIT,
        CONF_URL: git_file.git_url,
        CONF_FILE: git_file.filename,
    }
    if git_file.ref:
        conf[CONF_REF] = git_file.ref

    try:
        return MODEL_SOURCE_SCHEMA(conf)
    except cv.Invalid as e:
        raise cv.Invalid(
            f"Could not find file '{git_file.filename}' in the repository. Please make sure it exists."
        ) from e


MODEL_SOURCE_SCHEMA = cv.Any(
    _validate_source_shorthand,
    cv.typed_schema(
        {
            TYPE_GIT: GIT_SCHEMA,
            TYPE_LOCAL: LOCAL_SCHEMA,
            TYPE_HTTP: HTTP_SCHEMA,
        }
    ),
    msg="Not a valid model name, local path, http(s) url, or github shorthand",
)

MODEL_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_ID): cv.declare_id(WakeWordModel),
        cv.Optional(CONF_MODEL): MODEL_SOURCE_SCHEMA,
        cv.Optional(CONF_PROBABILITY_CUTOFF): cv.percentage,
        cv.Optional(CONF_SLIDING_WINDOW_SIZE): cv.positive_int,
        cv.Optional(CONF_INTERNAL, default=False): cv.boolean,
        cv.GenerateID(CONF_RAW_DATA_ID): cv.declare_id(cg.uint8),
    }
)

# Provides a default VAD model that could be overridden
VAD_MODEL_SCHEMA = MODEL_SCHEMA.extend(
    cv.Schema(
        {
            cv.Optional(
                CONF_MODEL,
                default="vad",
            ): MODEL_SOURCE_SCHEMA,
        }
    )
)


def _maybe_empty_vad_schema(value):
    # Idea borrowed from uart/__init__.py's ``maybe_empty_debug`` function. Accessed 2 July 2024.
    # Loads a default VAD model without any parameters overridden.
    if value is None:
        value = {}
    return VAD_MODEL_SCHEMA(value)


def _download_http_models(config: ConfigType) -> ConfigType:
    """Download every http-sourced manifest and model file in two concurrent
    batches (all manifests, then all model files).

    The model file's URL only becomes known once its manifest has been
    fetched and parsed, so the two stages cannot be merged into one batch.
    """
    model_parameters = [*config[CONF_MODELS]]
    if vad := config.get(CONF_VAD):
        model_parameters.append(vad)
    # Keyed by cache path so a URL referenced twice is fetched and parsed once
    http_models: dict[Path, str] = {
        _compute_local_file_path(model_config): model_config[CONF_URL]
        for parameters in model_parameters
        if (model_config := parameters.get(CONF_MODEL)) is not None
        and model_config.get(CONF_TYPE) == TYPE_HTTP
    }
    if not http_models:
        return config

    external_files.download_content_many(
        ((url, path / "manifest.json") for path, url in http_models.items()),
        description="wake word manifest(s)",
    )

    model_files: list[tuple[str, Path]] = []
    errors: list[cv.Invalid] = []
    for path, url in http_models.items():
        try:
            manifest_data = json.loads((path / "manifest.json").read_bytes())
        except (OSError, ValueError) as e:
            errors.append(cv.Invalid(f"Invalid manifest file at {url}: {e}"))
            continue
        if not isinstance(manifest_data, dict):
            errors.append(
                cv.Invalid(f"Manifest file at {url} must contain a JSON object")
            )
            continue
        model = manifest_data.get(CONF_MODEL)
        if not isinstance(model, str):
            errors.append(
                cv.Invalid(f"Manifest file at {url} is missing the 'model' key")
            )
            continue
        model_files.append((urljoin(url, model), path / model))
    if errors:
        raise cv.MultipleInvalid(errors)

    external_files.download_content_many(model_files, description="wake word model(s)")
    return config


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(MicroWakeWord),
            cv.Optional(
                CONF_MICROPHONE, default={}
            ): microphone.microphone_source_schema(
                min_bits_per_sample=16,
                max_bits_per_sample=16,
                min_channels=1,
                max_channels=1,
            ),
            cv.Required(CONF_MODELS): cv.ensure_list(
                cv.maybe_simple_value(MODEL_SCHEMA, key=CONF_MODEL)
            ),
            cv.Optional(CONF_ON_WAKE_WORD_DETECTED): automation.validate_automation(
                single=True
            ),
            cv.Optional(CONF_VAD): _maybe_empty_vad_schema,
            cv.Optional(CONF_STOP_AFTER_DETECTION, default=True): cv.boolean,
            cv.Optional(CONF_TASK_STACK_IN_PSRAM): psram.validate_task_stack_in_psram,
            cv.Optional(CONF_MODEL): cv.invalid(
                f"The {CONF_MODEL} parameter has moved to be a list element under the {CONF_MODELS} parameter."
            ),
            cv.Optional(CONF_PROBABILITY_CUTOFF): cv.invalid(
                f"The {CONF_PROBABILITY_CUTOFF} parameter has moved to be a list element under the {CONF_MODELS} parameter."
            ),
            cv.Optional(CONF_SLIDING_WINDOW_AVERAGE_SIZE): cv.invalid(
                f"The {CONF_SLIDING_WINDOW_AVERAGE_SIZE} parameter has been renamed to {CONF_SLIDING_WINDOW_SIZE} and moved to be a list element under the {CONF_MODELS} parameter."
            ),
        }
    ).extend(cv.COMPONENT_SCHEMA),
    cv.only_on_esp32,
    _download_http_models,
)


def _load_model_data(manifest_path: Path):
    with manifest_path.open(encoding="utf-8") as f:
        manifest = json.load(f)

    _validate_manifest_version(manifest)

    model_path = manifest_path.parent / manifest[CONF_MODEL]

    with model_path.open("rb") as f:
        model = f.read()

    if manifest.get(KEY_VERSION) == 1:
        manifest = _convert_manifest_v1_to_v2(manifest)

    return manifest, model


def _model_config_to_manifest_data(model_config):
    if model_config[CONF_TYPE] == TYPE_GIT:
        # compute path to model file
        key = f"{model_config[CONF_URL]}@{model_config.get(CONF_REF)}"
        base_dir = Path(CORE.data_dir) / DOMAIN
        h = hashlib.new("sha256")
        h.update(key.encode())
        file: Path = base_dir / h.hexdigest()[:8] / model_config[CONF_FILE]

    elif model_config[CONF_TYPE] == TYPE_LOCAL:
        file = Path(model_config[CONF_PATH])

    elif model_config[CONF_TYPE] == TYPE_HTTP:
        file = _compute_local_file_path(model_config) / "manifest.json"

    else:
        raise ValueError(f"Unsupported config type: {model_config[CONF_TYPE]}")

    return _load_model_data(file)


def _feature_step_size_validate(config):
    features_step_size = None

    for model_parameters in config[CONF_MODELS]:
        model_config = model_parameters.get(CONF_MODEL)
        manifest, _ = _model_config_to_manifest_data(model_config)

        model_step_size = manifest[KEY_MICRO][CONF_FEATURE_STEP_SIZE]

        if features_step_size is None:
            features_step_size = model_step_size
        elif features_step_size != model_step_size:
            raise cv.Invalid("Cannot load models with different features step sizes")


FINAL_VALIDATE_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.Required(
                CONF_MICROPHONE
            ): microphone.final_validate_microphone_source_schema(
                "micro_wake_word", sample_rate=16000
            ),
        },
        extra=cv.ALLOW_EXTRA,
    ),
    _feature_step_size_validate,
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    mic_source = await microphone.microphone_source_to_code(config[CONF_MICROPHONE])
    cg.add(var.set_microphone_source(mic_source))

    cg.add_define("USE_MICRO_WAKE_WORD")
    ota.request_ota_state_listeners()

    if config.get(CONF_TASK_STACK_IN_PSRAM):
        cg.add(var.set_task_stack_in_psram(True))
        psram.request_external_task_stack()

    esp32.add_idf_component(name="espressif/esp-tflite-micro", ref="1.3.3~1")
    # Pin esp-nn for stable future builds (esp-tflite-micro depends on esp-nn)
    esp32.add_idf_component(name="espressif/esp-nn", ref="1.1.2")

    esp32.add_idf_component(name="esphome/esp-micro-speech-features", ref="1.2.3")

    cg.add_build_flag("-DTF_LITE_STATIC_MEMORY")
    cg.add_build_flag("-DTF_LITE_DISABLE_X86_NEON")
    cg.add_build_flag("-DESP_NN")

    if vad_model := config.get(CONF_VAD):
        cg.add_define("USE_MICRO_WAKE_WORD_VAD")

        # Use the general model loading code for the VAD codegen
        config[CONF_MODELS].append(vad_model)

    for i, model_parameters in enumerate(config[CONF_MODELS]):
        model_config = model_parameters.get(CONF_MODEL)
        data = []
        manifest, data = _model_config_to_manifest_data(model_config)

        rhs = [HexInt(x) for x in data]
        prog_arr = cg.progmem_array(model_parameters[CONF_RAW_DATA_ID], rhs)

        probability_cutoff = model_parameters.get(
            CONF_PROBABILITY_CUTOFF, manifest[KEY_MICRO][CONF_PROBABILITY_CUTOFF]
        )
        quantized_probability_cutoff = int(probability_cutoff * 255)

        sliding_window_size = model_parameters.get(
            CONF_SLIDING_WINDOW_SIZE,
            manifest[KEY_MICRO][CONF_SLIDING_WINDOW_SIZE],
        )

        if manifest[KEY_WAKE_WORD] == "vad":
            cg.add(
                var.add_vad_model(
                    prog_arr,
                    quantized_probability_cutoff,
                    sliding_window_size,
                    manifest[KEY_MICRO][CONF_TENSOR_ARENA_SIZE],
                )
            )
        else:
            # Only enable the first wake word by default. After first boot, the enable state is saved/loaded to the flash
            default_enabled = i == 0
            wake_word_model = cg.new_Pvariable(
                model_parameters[CONF_ID],
                str(model_parameters[CONF_ID]),
                prog_arr,
                quantized_probability_cutoff,
                sliding_window_size,
                manifest[KEY_WAKE_WORD],
                manifest[KEY_MICRO][CONF_TENSOR_ARENA_SIZE],
                default_enabled,
                model_parameters[CONF_INTERNAL],
            )

            for lang in manifest[KEY_TRAINED_LANGUAGES]:
                cg.add(wake_word_model.add_trained_language(lang))

            cg.add(var.add_wake_word_model(wake_word_model))

    cg.add(var.set_features_step_size(manifest[KEY_MICRO][CONF_FEATURE_STEP_SIZE]))
    cg.add(var.set_stop_after_detection(config[CONF_STOP_AFTER_DETECTION]))

    if on_wake_word_detection_config := config.get(CONF_ON_WAKE_WORD_DETECTED):
        await automation.build_automation(
            var.get_wake_word_detected_trigger(),
            [(cg.std_string, "wake_word")],
            on_wake_word_detection_config,
        )


MICRO_WAKE_WORD_ACTION_SCHEMA = cv.Schema({cv.GenerateID(): cv.use_id(MicroWakeWord)})


@register_action(
    "micro_wake_word.start",
    StartAction,
    MICRO_WAKE_WORD_ACTION_SCHEMA,
    synchronous=True,
)
@register_action(
    "micro_wake_word.stop", StopAction, MICRO_WAKE_WORD_ACTION_SCHEMA, synchronous=True
)
@register_condition(
    "micro_wake_word.is_running", IsRunningCondition, MICRO_WAKE_WORD_ACTION_SCHEMA
)
async def micro_wake_word_action_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var


MICRO_WAKE_WORLD_MODEL_ACTION_SCHEMA = automation.maybe_simple_id(
    {
        cv.Required(CONF_ID): cv.use_id(WakeWordModel),
    }
)


@register_action(
    "micro_wake_word.enable_model",
    EnableModelAction,
    MICRO_WAKE_WORLD_MODEL_ACTION_SCHEMA,
    synchronous=True,
)
@register_action(
    "micro_wake_word.disable_model",
    DisableModelAction,
    MICRO_WAKE_WORLD_MODEL_ACTION_SCHEMA,
    synchronous=True,
)
@register_condition(
    "micro_wake_word.model_is_enabled",
    ModelIsEnabledCondition,
    MICRO_WAKE_WORLD_MODEL_ACTION_SCHEMA,
)
async def model_action(config, action_id, template_arg, args):
    parent = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, parent)



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import CONF_SOURCE_ID

from .. import Number, number_ns

NumberSensor = number_ns.class_("NumberSensor", sensor.Sensor, cg.Component)


CONFIG_SCHEMA = (
    sensor.sensor_schema(NumberSensor)
    .extend(
        {
            cv.Required(CONF_SOURCE_ID): cv.use_id(Number),
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
)


async def to_code(config):
    source = await cg.get_variable(config[CONF_SOURCE_ID])
    var = await sensor.new_sensor(config, source)
    await cg.register_component(var, config)



import esphome.codegen as cg

st7789v_ns = cg.esphome_ns.namespace("st7789v")

DEPRECATED_COMPONENT = """
The 'st7789v' component is deprecated and no new functionality will be added to it.
PRs should target the newer and more performant 'mipi_spi' component.
"""



from collections.abc import Callable
from dataclasses import dataclass
import logging
from typing import Any

from esphome import pins
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_IGNORE_PIN_VALIDATION_ERROR,
    CONF_IGNORE_STRAPPING_WARNING,
    CONF_INVERTED,
    CONF_MODE,
    CONF_NUMBER,
    CONF_OPEN_DRAIN,
    CONF_OUTPUT,
    PLATFORM_ESP32,
)
from esphome.core import CORE
from esphome.types import ConfigType

from . import boards
from .const import (
    KEY_BOARD,
    KEY_ESP32,
    KEY_VARIANT,
    VARIANT_ESP32,
    VARIANT_ESP32C2,
    VARIANT_ESP32C3,
    VARIANT_ESP32C5,
    VARIANT_ESP32C6,
    VARIANT_ESP32C61,
    VARIANT_ESP32H2,
    VARIANT_ESP32H4,
    VARIANT_ESP32H21,
    VARIANT_ESP32P4,
    VARIANT_ESP32S2,
    VARIANT_ESP32S3,
    VARIANT_ESP32S31,
    esp32_ns,
)
from .gpio_esp32 import esp32_validate_gpio_pin, esp32_validate_supports
from .gpio_esp32_c2 import esp32_c2_validate_gpio_pin, esp32_c2_validate_supports
from .gpio_esp32_c3 import esp32_c3_validate_gpio_pin, esp32_c3_validate_supports
from .gpio_esp32_c5 import esp32_c5_validate_gpio_pin, esp32_c5_validate_supports
from .gpio_esp32_c6 import esp32_c6_validate_gpio_pin, esp32_c6_validate_supports
from .gpio_esp32_c61 import esp32_c61_validate_gpio_pin, esp32_c61_validate_supports
from .gpio_esp32_h2 import esp32_h2_validate_gpio_pin, esp32_h2_validate_supports
from .gpio_esp32_h4 import esp32_h4_validate_gpio_pin, esp32_h4_validate_supports
from .gpio_esp32_h21 import esp32_h21_validate_gpio_pin, esp32_h21_validate_supports
from .gpio_esp32_p4 import esp32_p4_validate_gpio_pin, esp32_p4_validate_supports
from .gpio_esp32_s2 import esp32_s2_validate_gpio_pin, esp32_s2_validate_supports
from .gpio_esp32_s3 import (
    esp32_s3_final_validate_pins,
    esp32_s3_validate_gpio_pin,
    esp32_s3_validate_supports,
)
from .gpio_esp32_s31 import esp32_s31_validate_gpio_pin, esp32_s31_validate_supports

ESP32InternalGPIOPin = esp32_ns.class_("ESP32InternalGPIOPin", cg.InternalGPIOPin)


_LOGGER = logging.getLogger(__name__)


def _lookup_pin(value):
    board = CORE.data[KEY_ESP32][KEY_BOARD]
    board_pins = boards.ESP32_BOARD_PINS.get(board, {})

    # Resolved aliased board pins (shorthand when two boards have the same pin configuration)
    while isinstance(board_pins, str):
        board_pins = boards.ESP32_BOARD_PINS[board_pins]

    if value in board_pins:
        return board_pins[value]
    if value in boards.ESP32_BASE_PINS:
        return boards.ESP32_BASE_PINS[value]
    raise cv.Invalid(f"Cannot resolve pin name '{value}' for board {board}.")


def _translate_pin(value):
    if isinstance(value, dict) or value is None:
        raise cv.Invalid(
            "This variable only supports pin numbers, not full pin schemas (with inverted and mode)."
        )
    if isinstance(value, int) and not isinstance(value, bool):
        return value
    if not isinstance(value, str):
        raise cv.Invalid(f"Invalid pin number: {value}")
    try:
        return int(value)
    except ValueError:
        pass
    if value.startswith("GPIO"):
        return cv.int_(value[len("GPIO") :].strip())
    return _lookup_pin(value)


@dataclass
class ESP32ValidationFunctions:
    pin_validation: Callable[[int], int]
    usage_validation: Callable[[dict[str, Any]], dict[str, Any]]
    final_validate: Callable[[ConfigType], None] | None = None


_esp32_validations = {
    VARIANT_ESP32: ESP32ValidationFunctions(
        pin_validation=esp32_validate_gpio_pin, usage_validation=esp32_validate_supports
    ),
    VARIANT_ESP32C2: ESP32ValidationFunctions(
        pin_validation=esp32_c2_validate_gpio_pin,
        usage_validation=esp32_c2_validate_supports,
    ),
    VARIANT_ESP32C3: ESP32ValidationFunctions(
        pin_validation=esp32_c3_validate_gpio_pin,
        usage_validation=esp32_c3_validate_supports,
    ),
    VARIANT_ESP32C5: ESP32ValidationFunctions(
        pin_validation=esp32_c5_validate_gpio_pin,
        usage_validation=esp32_c5_validate_supports,
    ),
    VARIANT_ESP32C6: ESP32ValidationFunctions(
        pin_validation=esp32_c6_validate_gpio_pin,
        usage_validation=esp32_c6_validate_supports,
    ),
    VARIANT_ESP32C61: ESP32ValidationFunctions(
        pin_validation=esp32_c61_validate_gpio_pin,
        usage_validation=esp32_c61_validate_supports,
    ),
    VARIANT_ESP32H2: ESP32ValidationFunctions(
        pin_validation=esp32_h2_validate_gpio_pin,
        usage_validation=esp32_h2_validate_supports,
    ),
    VARIANT_ESP32H4: ESP32ValidationFunctions(
        pin_validation=esp32_h4_validate_gpio_pin,
        usage_validation=esp32_h4_validate_supports,
    ),
    VARIANT_ESP32H21: ESP32ValidationFunctions(
        pin_validation=esp32_h21_validate_gpio_pin,
        usage_validation=esp32_h21_validate_supports,
    ),
    VARIANT_ESP32P4: ESP32ValidationFunctions(
        pin_validation=esp32_p4_validate_gpio_pin,
        usage_validation=esp32_p4_validate_supports,
    ),
    VARIANT_ESP32S2: ESP32ValidationFunctions(
        pin_validation=esp32_s2_validate_gpio_pin,
        usage_validation=esp32_s2_validate_supports,
    ),
    VARIANT_ESP32S3: ESP32ValidationFunctions(
        pin_validation=esp32_s3_validate_gpio_pin,
        usage_validation=esp32_s3_validate_supports,
        final_validate=esp32_s3_final_validate_pins,
    ),
    VARIANT_ESP32S31: ESP32ValidationFunctions(
        pin_validation=esp32_s31_validate_gpio_pin,
        usage_validation=esp32_s31_validate_supports,
    ),
}


def gpio_pin_number_validator(value):
    value = _translate_pin(value)
    board = CORE.data[KEY_ESP32][KEY_BOARD]
    board_pins = boards.ESP32_BOARD_PINS.get(board, {})

    # Resolved aliased board pins (shorthand when two boards have the same pin configuration)
    while isinstance(board_pins, str):
        board_pins = boards.ESP32_BOARD_PINS[board_pins]

    if value in board_pins.values():
        return value

    variant = CORE.data[KEY_ESP32][KEY_VARIANT]
    if variant not in _esp32_validations:
        raise cv.Invalid(f"Unsupported ESP32 variant {variant}")

    return value


def validate_gpio_pin(pin):
    variant = CORE.data[KEY_ESP32][KEY_VARIANT]
    if variant not in _esp32_validations:
        raise cv.Invalid(f"Unsupported ESP32 variant {variant}")

    ignore_pin_validation_warning = pin[CONF_IGNORE_PIN_VALIDATION_ERROR]
    try:
        pin[CONF_NUMBER] = _esp32_validations[variant].pin_validation(pin[CONF_NUMBER])
    except cv.Invalid as exc:
        if not ignore_pin_validation_warning:
            raise

        _LOGGER.warning(
            "Ignoring validation error on pin %d; error: %s",
            pin[CONF_NUMBER],
            exc,
        )
    else:
        # `ignore_pin_validation_error` only suppresses an error raised by the
        # variant's pin_validation above (e.g. SPI flash/PSRAM pins, invalid pin
        # numbers). If that didn't raise, the option is a no-op -- warn so the
        # user can clean it up, but don't block the build.
        if ignore_pin_validation_warning:
            _LOGGER.warning(
                "GPIO%d has no validation errors to ignore; "
                "remove `ignore_pin_validation_error: true` from this pin.",
                pin[CONF_NUMBER],
            )

    return pin


def validate_supports(value):
    mode = value[CONF_MODE]
    is_output = mode[CONF_OUTPUT]
    is_open_drain = mode[CONF_OPEN_DRAIN]
    variant = CORE.data[KEY_ESP32][KEY_VARIANT]
    if variant not in _esp32_validations:
        raise cv.Invalid(f"Unsupported ESP32 variant {variant}")

    if is_open_drain and not is_output:
        raise cv.Invalid(
            "Open-drain only works with output mode", [CONF_MODE, CONF_OPEN_DRAIN]
        )

    return _esp32_validations[variant].usage_validation(value)


# https://docs.espressif.com/projects/esp-idf/en/v3.3.5/api-reference/peripherals/gpio.html#_CPPv416gpio_drive_cap_t
gpio_drive_cap_t = cg.global_ns.enum("gpio_drive_cap_t")
DRIVE_STRENGTHS = {
    5.0: gpio_drive_cap_t.GPIO_DRIVE_CAP_0,
    10.0: gpio_drive_cap_t.GPIO_DRIVE_CAP_1,
    20.0: gpio_drive_cap_t.GPIO_DRIVE_CAP_2,
    40.0: gpio_drive_cap_t.GPIO_DRIVE_CAP_3,
}
gpio_num_t = cg.global_ns.enum("gpio_num_t")

CONF_DRIVE_STRENGTH = "drive_strength"

ESP32_PIN_SCHEMA = cv.All(
    pins.gpio_base_schema(ESP32InternalGPIOPin, gpio_pin_number_validator).extend(
        {
            cv.Optional(CONF_IGNORE_PIN_VALIDATION_ERROR, default=False): cv.boolean,
            cv.Optional(CONF_IGNORE_STRAPPING_WARNING, default=False): cv.boolean,
            cv.Optional(CONF_DRIVE_STRENGTH, default="20mA"): cv.All(
                cv.float_with_unit("current", "mA", optional_unit=True),
                cv.enum(DRIVE_STRENGTHS),
            ),
        }
    ),
    validate_gpio_pin,
    validate_supports,
)


@pins.PIN_SCHEMA_REGISTRY.register(PLATFORM_ESP32, ESP32_PIN_SCHEMA)
async def esp32_pin_to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    num = config[CONF_NUMBER]
    cg.add(var.set_pin(getattr(gpio_num_t, f"GPIO_NUM_{num}")))
    # Only set if true to avoid bloating setup() function
    # (inverted bit in pin_flags_ bitfield is zero-initialized to false)
    if config[CONF_INVERTED]:
        cg.add(var.set_inverted(True))
    if CONF_DRIVE_STRENGTH in config:
        cg.add(var.set_drive_strength(config[CONF_DRIVE_STRENGTH]))
    cg.add(var.set_flags(pins.gpio_flags_expr(config[CONF_MODE])))
    return var


def final_validate_pins(full_config: ConfigType) -> None:
    """Run the active variant's pin final-validation, if it defines one."""
    funcs = _esp32_validations.get(CORE.data[KEY_ESP32][KEY_VARIANT])
    if funcs is not None and funcs.final_validate is not None:
        funcs.final_validate(full_config)



from esphome import pins
import esphome.codegen as cg
from esphome.components import spi, weikai
import esphome.config_validation as cv
from esphome.const import CONF_ID, CONF_INVERTED, CONF_MODE, CONF_NUMBER

CODEOWNERS = ["@DrCoolZic"]
DEPENDENCIES = ["spi"]
AUTO_LOAD = ["weikai", "weikai_spi"]
MULTI_CONF = True
CONF_WK2168_SPI = "wk2168_spi"

weikai_spi_ns = cg.esphome_ns.namespace("weikai_spi")
weikai_ns = cg.esphome_ns.namespace("weikai")
WeikaiComponentSPI = weikai_spi_ns.class_(
    "WeikaiComponentSPI", weikai.WeikaiComponent, spi.SPIDevice
)
WeikaiGPIOPin = weikai_ns.class_(
    "WeikaiGPIOPin", cg.GPIOPin, cg.Parented.template(WeikaiComponentSPI)
)

CONFIG_SCHEMA = cv.All(
    weikai.WKBASE_SCHEMA.extend(
        {cv.GenerateID(): cv.declare_id(WeikaiComponentSPI)}
    ).extend(spi.spi_device_schema()),
    weikai.check_channel_max_4,
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    cg.add(var.set_name(str(config[CONF_ID])))
    await weikai.register_weikai(var, config)
    await spi.register_spi_device(var, config)


WK2168_PIN_SCHEMA = cv.All(
    weikai.WEIKAI_PIN_SCHEMA.extend(
        {
            cv.GenerateID(): cv.declare_id(WeikaiGPIOPin),
            cv.Required(CONF_WK2168_SPI): cv.use_id(WeikaiComponentSPI),
        },
    ),
    weikai.validate_pin_mode,
)


@pins.PIN_SCHEMA_REGISTRY.register(CONF_WK2168_SPI, WK2168_PIN_SCHEMA)
async def sc16is75x_pin_to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    parent = await cg.get_variable(config[CONF_WK2168_SPI])
    cg.add(var.set_parent(parent))
    num = config[CONF_NUMBER]
    cg.add(var.set_pin(num))
    cg.add(var.set_inverted(config[CONF_INVERTED]))
    cg.add(var.set_flags(pins.gpio_flags_expr(config[CONF_MODE])))
    return var



from esphome import automation
import esphome.codegen as cg
from esphome.components import climate, output, sensor
import esphome.config_validation as cv
from esphome.const import CONF_HUMIDITY_SENSOR, CONF_ID, CONF_SENSOR

pid_ns = cg.esphome_ns.namespace("pid")
PIDClimate = pid_ns.class_("PIDClimate", climate.Climate, cg.Component)
PIDAutotuneAction = pid_ns.class_("PIDAutotuneAction", automation.Action)
PIDResetIntegralTermAction = pid_ns.class_(
    "PIDResetIntegralTermAction", automation.Action
)
PIDSetControlParametersAction = pid_ns.class_(
    "PIDSetControlParametersAction", automation.Action
)

CONF_DEFAULT_TARGET_TEMPERATURE = "default_target_temperature"

CONF_KP = "kp"
CONF_KI = "ki"
CONF_STARTING_INTEGRAL_TERM = "starting_integral_term"
CONF_KD = "kd"
CONF_CONTROL_PARAMETERS = "control_parameters"
CONF_COOL_OUTPUT = "cool_output"
CONF_HEAT_OUTPUT = "heat_output"
CONF_NOISEBAND = "noiseband"
CONF_POSITIVE_OUTPUT = "positive_output"
CONF_NEGATIVE_OUTPUT = "negative_output"
CONF_MIN_INTEGRAL = "min_integral"
CONF_MAX_INTEGRAL = "max_integral"
CONF_OUTPUT_AVERAGING_SAMPLES = "output_averaging_samples"
CONF_DERIVATIVE_AVERAGING_SAMPLES = "derivative_averaging_samples"

# Deadband parameters
CONF_DEADBAND_PARAMETERS = "deadband_parameters"
CONF_THRESHOLD_HIGH = "threshold_high"
CONF_THRESHOLD_LOW = "threshold_low"
CONF_DEADBAND_OUTPUT_AVERAGING_SAMPLES = "deadband_output_averaging_samples"
CONF_KP_MULTIPLIER = "kp_multiplier"
CONF_KI_MULTIPLIER = "ki_multiplier"
CONF_KD_MULTIPLIER = "kd_multiplier"

CONFIG_SCHEMA = cv.All(
    climate.climate_schema(PIDClimate).extend(
        {
            cv.Required(CONF_SENSOR): cv.use_id(sensor.Sensor),
            cv.Optional(CONF_HUMIDITY_SENSOR): cv.use_id(sensor.Sensor),
            cv.Required(CONF_DEFAULT_TARGET_TEMPERATURE): cv.temperature,
            cv.Optional(CONF_COOL_OUTPUT): cv.use_id(output.FloatOutput),
            cv.Optional(CONF_HEAT_OUTPUT): cv.use_id(output.FloatOutput),
            cv.Optional(CONF_DEADBAND_PARAMETERS): cv.Schema(
                {
                    cv.Required(CONF_THRESHOLD_HIGH): cv.temperature_delta,
                    cv.Required(CONF_THRESHOLD_LOW): cv.temperature_delta,
                    cv.Optional(CONF_KP_MULTIPLIER, default=0.1): cv.float_,
                    cv.Optional(CONF_KI_MULTIPLIER, default=0.0): cv.float_,
                    cv.Optional(CONF_KD_MULTIPLIER, default=0.0): cv.float_,
                    cv.Optional(
                        CONF_DEADBAND_OUTPUT_AVERAGING_SAMPLES, default=1
                    ): cv.positive_not_null_int,
                }
            ),
            cv.Required(CONF_CONTROL_PARAMETERS): cv.Schema(
                {
                    cv.Required(CONF_KP): cv.float_,
                    cv.Optional(CONF_KI, default=0.0): cv.float_,
                    cv.Optional(CONF_KD, default=0.0): cv.float_,
                    cv.Optional(CONF_STARTING_INTEGRAL_TERM, default=0.0): cv.float_,
                    cv.Optional(CONF_MIN_INTEGRAL, default=-1): cv.float_,
                    cv.Optional(CONF_MAX_INTEGRAL, default=1): cv.float_,
                    cv.Optional(
                        CONF_DERIVATIVE_AVERAGING_SAMPLES, default=1
                    ): cv.positive_not_null_int,
                    cv.Optional(
                        CONF_OUTPUT_AVERAGING_SAMPLES, default=1
                    ): cv.positive_not_null_int,
                }
            ),
        }
    ),
    cv.has_at_least_one_key(CONF_COOL_OUTPUT, CONF_HEAT_OUTPUT),
)


async def to_code(config):
    var = await climate.new_climate(config)
    await cg.register_component(var, config)

    sens = await cg.get_variable(config[CONF_SENSOR])
    cg.add(var.set_sensor(sens))

    if CONF_HUMIDITY_SENSOR in config:
        sens = await cg.get_variable(config[CONF_HUMIDITY_SENSOR])
        cg.add(var.set_humidity_sensor(sens))

    if CONF_COOL_OUTPUT in config:
        out = await cg.get_variable(config[CONF_COOL_OUTPUT])
        cg.add(var.set_cool_output(out))
    if CONF_HEAT_OUTPUT in config:
        out = await cg.get_variable(config[CONF_HEAT_OUTPUT])
        cg.add(var.set_heat_output(out))
    params = config[CONF_CONTROL_PARAMETERS]
    cg.add(var.set_kp(params[CONF_KP]))
    cg.add(var.set_ki(params[CONF_KI]))
    cg.add(var.set_kd(params[CONF_KD]))
    cg.add(var.set_starting_integral_term(params[CONF_STARTING_INTEGRAL_TERM]))
    cg.add(var.set_derivative_samples(params[CONF_DERIVATIVE_AVERAGING_SAMPLES]))

    output_samples = params[CONF_OUTPUT_AVERAGING_SAMPLES]
    cg.add(var.set_output_samples(output_samples))

    if CONF_MIN_INTEGRAL in params:
        cg.add(var.set_min_integral(params[CONF_MIN_INTEGRAL]))
    if CONF_MAX_INTEGRAL in params:
        cg.add(var.set_max_integral(params[CONF_MAX_INTEGRAL]))

    deadband_output_samples = 1
    if CONF_DEADBAND_PARAMETERS in config:
        params = config[CONF_DEADBAND_PARAMETERS]
        cg.add(var.set_threshold_low(params[CONF_THRESHOLD_LOW]))
        cg.add(var.set_threshold_high(params[CONF_THRESHOLD_HIGH]))
        cg.add(var.set_kp_multiplier(params[CONF_KP_MULTIPLIER]))
        cg.add(var.set_ki_multiplier(params[CONF_KI_MULTIPLIER]))
        cg.add(var.set_kd_multiplier(params[CONF_KD_MULTIPLIER]))
        deadband_output_samples = params[CONF_DEADBAND_OUTPUT_AVERAGING_SAMPLES]
        cg.add(var.set_deadband_output_samples(deadband_output_samples))

    # Single shared output buffer sized to max of both modes
    cg.add(var.init_output_buffer(max(output_samples, deadband_output_samples)))

    cg.add(var.set_default_target_temperature(config[CONF_DEFAULT_TARGET_TEMPERATURE]))


@automation.register_action(
    "climate.pid.reset_integral_term",
    PIDResetIntegralTermAction,
    automation.maybe_simple_id(
        {
            cv.Required(CONF_ID): cv.use_id(PIDClimate),
        }
    ),
    synchronous=True,
)
async def pid_reset_integral_term(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)


@automation.register_action(
    "climate.pid.autotune",
    PIDAutotuneAction,
    automation.maybe_simple_id(
        {
            cv.Required(CONF_ID): cv.use_id(PIDClimate),
            cv.Optional(CONF_NOISEBAND, default=0.25): cv.float_,
            cv.Optional(
                CONF_POSITIVE_OUTPUT, default=1.0
            ): cv.possibly_negative_percentage,
            cv.Optional(
                CONF_NEGATIVE_OUTPUT, default=-1.0
            ): cv.possibly_negative_percentage,
        }
    ),
    synchronous=True,
)
async def esp8266_set_frequency_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)
    cg.add(var.set_noiseband(config[CONF_NOISEBAND]))
    cg.add(var.set_positive_output(config[CONF_POSITIVE_OUTPUT]))
    cg.add(var.set_negative_output(config[CONF_NEGATIVE_OUTPUT]))
    return var


@automation.register_action(
    "climate.pid.set_control_parameters",
    PIDSetControlParametersAction,
    automation.maybe_simple_id(
        {
            cv.Required(CONF_ID): cv.use_id(PIDClimate),
            cv.Required(CONF_KP): cv.templatable(cv.float_),
            cv.Optional(CONF_KI, default=0.0): cv.templatable(cv.float_),
            cv.Optional(CONF_KD, default=0.0): cv.templatable(cv.float_),
        }
    ),
    synchronous=True,
)
async def set_control_parameters(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)

    kp_template_ = await cg.templatable(config[CONF_KP], args, cg.float_)
    cg.add(var.set_kp(kp_template_))

    ki_template_ = await cg.templatable(config[CONF_KI], args, cg.float_)
    cg.add(var.set_ki(ki_template_))

    kd_template_ = await cg.templatable(config[CONF_KD], args, cg.float_)
    cg.add(var.set_kd(kd_template_))

    return var



"""Analyzer for RAM-stored strings in ESP8266/ESP32 firmware ELF files.

This module identifies strings that are stored in RAM sections (.data, .bss, .rodata)
rather than in flash sections (.irom0.text, .irom.text), which is important for
memory-constrained platforms like ESP8266.
"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
import logging
from pathlib import Path
import re
import subprocess

from .demangle import batch_demangle
from .toolchain import find_tool

_LOGGER = logging.getLogger(__name__)

# ESP8266: .rodata is in RAM (DRAM), not flash
# ESP32: .rodata is in flash, mapped to data bus
ESP8266_RAM_SECTIONS = frozenset([".data", ".rodata", ".bss"])
ESP8266_FLASH_SECTIONS = frozenset([".irom0.text", ".irom.text", ".text"])

# ESP32: .rodata is memory-mapped from flash
ESP32_RAM_SECTIONS = frozenset([".data", ".bss", ".dram0.data", ".dram0.bss"])
ESP32_FLASH_SECTIONS = frozenset([".text", ".rodata", ".flash.text", ".flash.rodata"])

# nm symbol types for data symbols (D=global data, d=local data, R=rodata, B=bss)
DATA_SYMBOL_TYPES = frozenset(["D", "d", "R", "r", "B", "b"])


@dataclass
class SectionInfo:
    """Information about an ELF section."""

    name: str
    address: int
    size: int


@dataclass
class RamString:
    """A string found in RAM."""

    section: str
    address: int
    content: str

    @property
    def size(self) -> int:
        """Size in bytes including null terminator."""
        return len(self.content) + 1


@dataclass
class RamSymbol:
    """A symbol found in RAM."""

    name: str
    sym_type: str
    address: int
    size: int
    section: str
    demangled: str = ""  # Demangled name, set after batch demangling
    aliases: list[str] = field(default_factory=list)  # Other names at same address


class RamStringsAnalyzer:
    """Analyzes ELF files to find strings stored in RAM."""

    def __init__(
        self,
        elf_path: str,
        objdump_path: str | None = None,
        min_length: int = 8,
        platform: str = "esp32",
    ) -> None:
        """Initialize the RAM strings analyzer.

        Args:
            elf_path: Path to the ELF file to analyze
            objdump_path: Path to objdump binary (used to find other tools)
            min_length: Minimum string length to report (default: 8)
            platform: Platform name ("esp8266", "esp32", etc.) for section mapping
        """
        self.elf_path = Path(elf_path)
        if not self.elf_path.exists():
            raise FileNotFoundError(f"ELF file not found: {elf_path}")

        self.objdump_path = objdump_path
        self.min_length = min_length
        self.platform = platform

        # Set RAM/flash sections based on platform
        if self.platform == "esp8266":
            self.ram_sections = ESP8266_RAM_SECTIONS
            self.flash_sections = ESP8266_FLASH_SECTIONS
        else:
            # ESP32 and other platforms
            self.ram_sections = ESP32_RAM_SECTIONS
            self.flash_sections = ESP32_FLASH_SECTIONS

        self.sections: dict[str, SectionInfo] = {}
        self.ram_strings: list[RamString] = []
        self.ram_symbols: list[RamSymbol] = []

    def _run_command(self, cmd: list[str]) -> str:
        """Run a command and return its output."""
        try:
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
            return result.stdout
        except subprocess.CalledProcessError as e:
            _LOGGER.debug("Command failed: %s - %s", " ".join(cmd), e.stderr)
            raise
        except FileNotFoundError:
            _LOGGER.warning("Command not found: %s", cmd[0])
            raise

    def analyze(self) -> None:
        """Perform the full RAM analysis."""
        self._parse_sections()
        self._extract_strings()
        self._analyze_symbols()
        self._demangle_symbols()

    def _parse_sections(self) -> None:
        """Parse section headers from ELF file."""
        objdump = find_tool("objdump", self.objdump_path)
        if not objdump:
            _LOGGER.error("Could not find objdump command")
            return

        try:
            output = self._run_command([objdump, "-h", str(self.elf_path)])
        except (subprocess.CalledProcessError, FileNotFoundError):
            return

        # Parse section headers
        # Format: Idx Name          Size      VMA       LMA       File off  Algn
        section_pattern = re.compile(
            r"^\s*\d+\s+(\S+)\s+([0-9a-fA-F]+)\s+([0-9a-fA-F]+)"
        )

        for line in output.split("\n"):
            if match := section_pattern.match(line):
                name = match.group(1)
                size = int(match.group(2), 16)
                vma = int(match.group(3), 16)
                self.sections[name] = SectionInfo(name, vma, size)

    def _extract_strings(self) -> None:
        """Extract strings from RAM sections."""
        objdump = find_tool("objdump", self.objdump_path)
        if not objdump:
            return

        for section_name in self.ram_sections:
            if section_name not in self.sections:
                continue

            try:
                output = self._run_command(
                    [objdump, "-s", "-j", section_name, str(self.elf_path)]
                )
            except subprocess.CalledProcessError:
                # Section may exist but have no content (e.g., .bss)
                continue
            except FileNotFoundError:
                continue

            strings = self._parse_hex_dump(output, section_name)
            self.ram_strings.extend(strings)

    def _parse_hex_dump(self, output: str, section_name: str) -> list[RamString]:
        """Parse hex dump output to extract strings.

        Args:
            output: Output from objdump -s
            section_name: Name of the section being parsed

        Returns:
            List of RamString objects
        """
        strings: list[RamString] = []
        current_string = bytearray()
        string_start_addr = 0

        for line in output.split("\n"):
            # Lines look like: " 3ffef8a0 00000000 00000000 00000000 00000000  ................"
            match = re.match(r"^\s+([0-9a-fA-F]+)\s+((?:[0-9a-fA-F]{2,8}\s*)+)", line)
            if not match:
                continue

            addr = int(match.group(1), 16)
            hex_data = match.group(2).strip()

            # Convert hex to bytes
            hex_bytes = hex_data.split()
            byte_offset = 0
            for hex_chunk in hex_bytes:
                # Handle both byte-by-byte and word formats
                for i in range(0, len(hex_chunk), 2):
                    byte_val = int(hex_chunk[i : i + 2], 16)
                    if 0x20 <= byte_val <= 0x7E:  # Printable ASCII
                        if not current_string:
                            string_start_addr = addr + byte_offset
                        current_string.append(byte_val)
                    else:
                        if byte_val == 0 and len(current_string) >= self.min_length:
                            # Found null terminator
                            strings.append(
                                RamString(
                                    section=section_name,
                                    address=string_start_addr,
                                    content=current_string.decode(
                                        "ascii", errors="ignore"
                                    ),
                                )
                            )
                        current_string = bytearray()
                    byte_offset += 1

        return strings

    def _analyze_symbols(self) -> None:
        """Analyze symbols in RAM sections."""
        nm = find_tool("nm", self.objdump_path)
        if not nm:
            return

        try:
            output = self._run_command([nm, "-S", "--size-sort", str(self.elf_path)])
        except (subprocess.CalledProcessError, FileNotFoundError):
            return

        # Track symbols by address so aliases (multiple names for the same
        # object, e.g. the newlib __lock___* mutexes that all alias one
        # StaticSemaphore_t) are reported once instead of once per name.
        symbols_by_addr: dict[int, RamSymbol] = {}

        for line in output.split("\n"):
            parts = line.split()
            if len(parts) < 4:
                continue

            try:
                addr = int(parts[0], 16)
                size = int(parts[1], 16) if parts[1] != "?" else 0
            except ValueError:
                continue

            sym_type = parts[2]
            name = " ".join(parts[3:])

            # Filter for data symbols
            if sym_type not in DATA_SYMBOL_TYPES:
                continue

            if (existing := symbols_by_addr.get(addr)) is not None:
                # Prefer a global (uppercase type) name as the primary so
                # nm output order can't hide it behind a local alias.
                if sym_type.isupper() and existing.sym_type.islower():
                    existing.aliases.append(existing.name)
                    existing.name = name
                    existing.sym_type = sym_type
                else:
                    existing.aliases.append(name)
                existing.size = max(existing.size, size)
                continue

            # Check if symbol is in a RAM section
            for section_name in self.ram_sections:
                if section_name not in self.sections:
                    continue

                section = self.sections[section_name]
                if section.address <= addr < section.address + section.size:
                    symbol = RamSymbol(
                        name=name,
                        sym_type=sym_type,
                        address=addr,
                        size=size,
                        section=section_name,
                    )
                    symbols_by_addr[addr] = symbol
                    self.ram_symbols.append(symbol)
                    break

    def _demangle_symbols(self) -> None:
        """Batch demangle all RAM symbol names."""
        if not self.ram_symbols:
            return

        # Collect all symbol names and demangle them
        symbol_names = [s.name for s in self.ram_symbols]
        demangle_cache = batch_demangle(symbol_names, objdump_path=self.objdump_path)

        # Assign demangled names to symbols
        for symbol in self.ram_symbols:
            symbol.demangled = demangle_cache.get(symbol.name, symbol.name)

    def _get_sections_size(self, section_names: frozenset[str]) -> int:
        """Get total size of specified sections."""
        return sum(
            section.size
            for name, section in self.sections.items()
            if name in section_names
        )

    def get_total_ram_usage(self) -> int:
        """Get total RAM usage from RAM sections."""
        return self._get_sections_size(self.ram_sections)

    def get_total_flash_usage(self) -> int:
        """Get total flash usage from flash sections."""
        return self._get_sections_size(self.flash_sections)

    def get_total_string_bytes(self) -> int:
        """Get total bytes used by strings in RAM."""
        return sum(s.size for s in self.ram_strings)

    def get_repeated_strings(self) -> list[tuple[str, int]]:
        """Find strings that appear multiple times.

        Returns:
            List of (string, count) tuples sorted by potential savings
        """
        string_counts: dict[str, int] = defaultdict(int)
        for ram_string in self.ram_strings:
            string_counts[ram_string.content] += 1

        return sorted(
            [(s, c) for s, c in string_counts.items() if c > 1],
            key=lambda x: x[1] * (len(x[0]) + 1),
            reverse=True,
        )

    def get_long_strings(self, min_len: int = 20) -> list[RamString]:
        """Get strings longer than the specified length.

        Args:
            min_len: Minimum string length

        Returns:
            List of RamString objects sorted by length
        """
        return sorted(
            [s for s in self.ram_strings if len(s.content) >= min_len],
            key=lambda x: len(x.content),
            reverse=True,
        )

    def get_largest_symbols(self, min_size: int = 100) -> list[RamSymbol]:
        """Get RAM symbols larger than the specified size.

        Args:
            min_size: Minimum symbol size in bytes

        Returns:
            List of RamSymbol objects sorted by size
        """
        return sorted(
            [s for s in self.ram_symbols if s.size >= min_size],
            key=lambda x: x.size,
            reverse=True,
        )

    def generate_report(self, show_all_sections: bool = False) -> str:
        """Generate a formatted RAM strings analysis report.

        Args:
            show_all_sections: If True, show all sections, not just RAM

        Returns:
            Formatted report string
        """
        lines: list[str] = []
        table_width = 80

        lines.append("=" * table_width)
        lines.append(
            f"RAM Strings Analysis ({self.platform.upper()})".center(table_width)
        )
        lines.append("=" * table_width)
        lines.append("")

        # Section Analysis
        lines.append("SECTION ANALYSIS")
        lines.append("-" * table_width)
        lines.append(f"{'Section':<20} {'Address':<12} {'Size':<12} {'Location'}")
        lines.append("-" * table_width)

        total_ram_usage = 0
        total_flash_usage = 0

        for name, section in sorted(self.sections.items(), key=lambda x: x[1].address):
            if name in self.ram_sections:
                location = "RAM"
                total_ram_usage += section.size
            elif name in self.flash_sections:
                location = "FLASH"
                total_flash_usage += section.size
            else:
                location = "OTHER"

            if show_all_sections or name in self.ram_sections:
                lines.append(
                    f"{name:<20} 0x{section.address:08x}   {section.size:>8} B   {location}"
                )

        lines.append("-" * table_width)
        lines.append(f"Total RAM sections size: {total_ram_usage:,} bytes")
        lines.append(f"Total Flash sections size: {total_flash_usage:,} bytes")

        # Strings in RAM
        lines.append("")
        lines.append("=" * table_width)
        lines.append("STRINGS IN RAM SECTIONS")
        lines.append("=" * table_width)
        lines.append(
            "Note: .bss sections contain uninitialized data (no strings to extract)"
        )

        # Group strings by section
        strings_by_section: dict[str, list[RamString]] = defaultdict(list)
        for ram_string in self.ram_strings:
            strings_by_section[ram_string.section].append(ram_string)

        for section_name in sorted(strings_by_section.keys()):
            section_strings = strings_by_section[section_name]
            lines.append(f"\nSection: {section_name}")
            lines.append("-" * 40)
            for ram_string in sorted(section_strings, key=lambda x: x.address):
                clean_string = ram_string.content[:100] + (
                    "..." if len(ram_string.content) > 100 else ""
                )
                lines.append(
                    f'  0x{ram_string.address:08x}: "{clean_string}" (len={len(ram_string.content)})'
                )

        # Large RAM symbols
        lines.append("")
        lines.append("=" * table_width)
        lines.append("LARGE DATA SYMBOLS IN RAM (>= 50 bytes)")
        lines.append("=" * table_width)

        largest_symbols = self.get_largest_symbols(50)
        lines.append(f"\n{'Symbol':<50} {'Type':<6} {'Size':<10} {'Section'}")
        lines.append("-" * table_width)

        for symbol in largest_symbols:
            # Use demangled name if available, otherwise raw name
            display_name = symbol.demangled or symbol.name
            # Truncate the name, not the alias note, so merged aliases stay
            # visible even for long demangled C++ names.
            alias_note = f" (+{len(symbol.aliases)} aliases)" if symbol.aliases else ""
            max_name_len = 49 - len(alias_note)
            if len(display_name) > max_name_len:
                display_name = display_name[:max_name_len]
            name_display = display_name + alias_note
            lines.append(
                f"{name_display:<50} {symbol.sym_type:<6} {symbol.size:>8} B  {symbol.section}"
            )

        # Summary
        lines.append("")
        lines.append("=" * table_width)
        lines.append("SUMMARY")
        lines.append("=" * table_width)
        lines.append(f"Total strings found in RAM: {len(self.ram_strings)}")
        total_string_bytes = self.get_total_string_bytes()
        lines.append(f"Total bytes used by strings: {total_string_bytes:,}")

        # Optimization targets
        lines.append("")
        lines.append("=" * table_width)
        lines.append("POTENTIAL OPTIMIZATION TARGETS")
        lines.append("=" * table_width)

        # Repeated strings
        repeated = self.get_repeated_strings()[:10]
        if repeated:
            lines.append("\nRepeated strings (could be deduplicated):")
            for string, count in repeated:
                savings = (count - 1) * (len(string) + 1)
                clean_string = string[:50] + ("..." if len(string) > 50 else "")
                lines.append(
                    f'  "{clean_string}" - appears {count} times (potential savings: {savings} bytes)'
                )

        # Long strings - platform-specific advice
        long_strings = self.get_long_strings(20)[:10]
        if long_strings:
            if self.platform == "esp8266":
                lines.append(
                    "\nLong strings that could be moved to PROGMEM (>= 20 chars):"
                )
            else:
                # ESP32: strings in DRAM are typically there for a reason
                # (interrupt handlers, pre-flash-init code, etc.)
                lines.append("\nLong strings in DRAM (>= 20 chars):")
                lines.append(
                    "Note: ESP32 DRAM strings may be required for interrupt/early-boot contexts"
                )
            for ram_string in long_strings:
                clean_string = ram_string.content[:60] + (
                    "..." if len(ram_string.content) > 60 else ""
                )
                lines.append(
                    f'  {ram_string.section} @ 0x{ram_string.address:08x}: "{clean_string}" ({len(ram_string.content)} bytes)'
                )

        lines.append("")
        return "\n".join(lines)



CODEOWNERS = ["@clydebarrow"]

CONF_DRAW_FROM_ORIGIN = "draw_from_origin"

DEPRECATED_COMPONENT = """
The 'qspi_dbi' component is deprecated and no new models will be added to it.
New model PRs should target the newer and more performant 'mipi_spi' component.
"""



from esphome.components.esp32 import add_idf_sdkconfig_option
import esphome.config_validation as cv

CODEOWNERS = ["@dentra"]

CONFIG_SCHEMA = cv.All(
    cv.Schema({}),
    cv.only_on_esp32,
)


async def to_code(config):
    # Increase the maximum supported size of headers section in HTTP request packet to be processed by the server
    add_idf_sdkconfig_option("CONFIG_HTTPD_MAX_REQ_HDR_LEN", 1024)



from collections.abc import Callable
import re

from esphome import automation
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import CONF_ID, CONF_ON_ERROR, CONF_ON_SUCCESS
from esphome.cpp_generator import MockObj, MockObjClass
from esphome.helpers import fnv1_hash_object_id

CODEOWNERS = ["@esphome/core"]

DOMAIN = "motion"
IS_PLATFORM_COMPONENT = True

#  C++ namespace / class
motion_ns = cg.esphome_ns.namespace("motion")
MotionComponent = motion_ns.class_("MotionComponent", cg.PollingComponent)

AXES = ["x", "y", "z"]

CONF_AXIS_MAP = "axis_map"
CONF_MOTION_ID = "motion_id"
CONF_TRANSFORM_MATRIX = "transform_matrix"

CalibrateLevelAction = motion_ns.class_("CalibrateLevelAction", automation.Action)
CalibrateHeadingAction = motion_ns.class_("CalibrateHeadingAction", automation.Action)
ClearCalibrationAction = motion_ns.class_("ClearCalibrationAction", automation.Action)

KEY_ACCELEROMETER = "accelerometer"
KEY_GYROSCOPE = "gyroscope"

SENSOR_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_MOTION_ID): cv.use_id(MotionComponent),
    }
)

_AXIS_REGEX = re.compile(r"^[+-]?[xyz]$", re.IGNORECASE)


def _axis_map(config: dict) -> dict:
    errors = []
    for key, axis in config.items():
        if _AXIS_REGEX.fullmatch(axis) is None:
            errors.append(
                cv.Invalid(
                    "Each 'axis_map' config value must be one of 'x', 'y' or 'z' (optionally preceded by '+' or '-').",
                    path=[key],
                )
            )
    values = {x.lower().removeprefix("-").removeprefix("+") for x in config.values()}
    if values != set(AXES):
        errors.append(cv.Invalid("Each axis may be mapped only once"))
    if errors:
        raise cv.MultipleInvalid(errors)
    return config


def _axis_map_to_matrix(config: dict[str, str]) -> list[float]:
    matrix = []
    for target_axis in AXES:
        source_axis = config[target_axis].lower()
        sign = -1.0 if source_axis.startswith("-") else 1.0
        source_axis = source_axis.removeprefix("+").removeprefix("-")

        row = [0.0, 0.0, 0.0]
        row[AXES.index(source_axis)] = sign
        matrix.extend(row)

    return matrix


def _transform_matrix(value):
    """Accept a flat list of 9 floats or a 3x3 nested list."""
    if not isinstance(value, list) or len(value) == 0:
        raise cv.Invalid("Expected a list of 9 numbers or a 3x3 nested list")
    # Nested 3x3
    if isinstance(value[0], list):
        if len(value) != 3:
            raise cv.Invalid(f"3x3 matrix must have 3 rows, got {len(value)}")
        flat = []
        for i, row in enumerate(value):
            if not isinstance(row, list) or len(row) != 3:
                raise cv.Invalid("Each row must be a list of 3 numbers", path=[i])
            flat.extend(cv.float_(v) for v in row)
        return flat
    # Flat list
    if len(value) != 9:
        raise cv.Invalid(f"Flat matrix must have exactly 9 values, got {len(value)}")
    return [cv.float_(v) for v in value]


def _validate_matrix_options(config):
    if CONF_AXIS_MAP in config and CONF_TRANSFORM_MATRIX in config:
        raise cv.Invalid(
            f"'{CONF_AXIS_MAP}' and '{CONF_TRANSFORM_MATRIX}' are mutually exclusive"
        )
    return config


#  Top-level CONFIG_SCHEMA
_CONFIG_SCHEMA = (
    cv.Schema(
        {
            cv.Optional(CONF_AXIS_MAP): cv.All(
                {cv.Required(k): cv.string_strict for k in AXES},
                _axis_map,
            ),
            cv.Optional(CONF_TRANSFORM_MATRIX): _transform_matrix,
        }
    )
    .extend(cv.polling_component_schema("250ms"))
    .add_extra(_validate_matrix_options)
)


def _add_data(has_accel: bool, has_gyro: bool) -> Callable[[dict], dict]:

    def validator(config):
        config = config.copy()
        config[KEY_ACCELEROMETER] = has_accel
        config[KEY_GYROSCOPE] = has_gyro
        return config

    return validator


def motion_schema(class_: MockObjClass, has_accel: bool, has_gyro: bool) -> cv.Schema:
    return _CONFIG_SCHEMA.extend(
        {
            cv.GenerateID(): cv.declare_id(class_),
        }
    ).add_extra(_add_data(has_accel, has_gyro))


#  Code generation
async def register_motion_component(var: MockObj, config) -> None:
    await cg.register_component(var, config)
    # Set preference key for NVS save/restore (based on component ID)
    obj_id = config[CONF_ID].id
    pref_hash = fnv1_hash_object_id(obj_id)
    cg.add(var.set_calibration_key(pref_hash))
    if axis_map := config.get(CONF_AXIS_MAP):
        cg.add(var.set_matrix(_axis_map_to_matrix(axis_map)))
    elif transform_matrix := config.get(CONF_TRANSFORM_MATRIX):
        cg.add(var.set_matrix(transform_matrix))


async def new_motion_component(config: dict) -> MockObj:
    var = cg.new_Pvariable(config[CONF_ID])
    await register_motion_component(var, config)
    return var


# --- Actions ---

CONF_SAVE = "save"

CALIBRATE_ACTION_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.use_id(MotionComponent),
        cv.Optional(CONF_SAVE, default=False): cv.boolean,
        cv.Optional(CONF_ON_SUCCESS): automation.validate_automation(single=True),
        cv.Optional(CONF_ON_ERROR): automation.validate_automation(single=True),
    }
)


async def _build_calibrate_action(config, action_id, template_arg, args):
    parent = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, parent)
    if config.get(CONF_SAVE):
        cg.add(var.set_save(True))
    if on_success := config.get(CONF_ON_SUCCESS):
        await automation.build_automation(var.get_success_trigger(), [], on_success)
    if on_error := config.get(CONF_ON_ERROR):
        await automation.build_automation(var.get_error_trigger(), [], on_error)
    return var


@automation.register_action(
    "motion.calibrate_level",
    CalibrateLevelAction,
    CALIBRATE_ACTION_SCHEMA,
    synchronous=True,
)
async def calibrate_level_to_code(config, action_id, template_arg, args):
    return await _build_calibrate_action(config, action_id, template_arg, args)


@automation.register_action(
    "motion.calibrate_heading",
    CalibrateHeadingAction,
    CALIBRATE_ACTION_SCHEMA,
    synchronous=True,
)
async def calibrate_heading_to_code(config, action_id, template_arg, args):
    return await _build_calibrate_action(config, action_id, template_arg, args)


CLEAR_ACTION_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.use_id(MotionComponent),
        cv.Optional(CONF_SAVE, default=False): cv.boolean,
    }
)


@automation.register_action(
    "motion.clear_calibration",
    ClearCalibrationAction,
    CLEAR_ACTION_SCHEMA,
    synchronous=True,
)
async def clear_calibration_to_code(config, action_id, template_arg, args):
    parent = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, parent)
    if config.get(CONF_SAVE):
        cg.add(var.set_save(True))
    return var



"""Print the ESP-IDF version of a given framework root.

Run via ``python <this file> <idf_framework_root>``. PYTHONPATH must include
``<idf_framework_root>/tools`` so ``idf_tools`` is importable.
"""

# pylint: disable=import-error  # idf_tools is on PYTHONPATH at runtime only

import sys

from idf_tools import g, get_idf_version

g.idf_path = sys.argv[1]
print(get_idf_version())
